import torch
from torch import nn
import torch.nn.functional as F
from .utils import *
from .loadData import *
import os
import json
import numpy as np
from torch. utils.data import DataLoader, Dataset, TensorDataset
import glob
import logging
logging.getLogger().setLevel(logging.INFO)


#############################
# 2025.01.08 adjust for istar
#############################
def build_loaders(batch_size, image_embed_path, spatial_pos_path, reduced_mtx_path, 
                  image_clacss='HIPT', dataset_class='Visium'):
    setup_seed(666)

    ## set image_paths, according 'image_clacss'
    if image_clacss == 'istar':
    
        dataset = DatasetCreat_istar(
            image_paths=image_embed_path,
            spatial_pos_path=spatial_pos_path,
            reduced_mtx_path=reduced_mtx_path,   
            dataset_class=dataset_class
        )
    elif image_clacss == 'HIPT' or image_clacss == 'Virchow2':
        import glob
        image_paths = glob.glob(str(image_embed_path))
        image_paths.sort()

        dataset = DatasetCreat(
            image_paths=image_paths,
            spatial_pos_path=spatial_pos_path,
            reduced_mtx_path=reduced_mtx_path,   
            dataset_class=dataset_class
        )
    else:
        raise ValueError('Invalid image_class. Only "istar", "HIPT" and "Virchow2" are supported.')    

    train_size = int(0.8 * len(dataset))
    test_size = len(dataset) - train_size
    train_dataset, test_dataset = torch.utils.data.random_split(dataset, [train_size, test_size])
    print("train/test split completed")
    print(len(train_dataset), len(test_dataset))

    ## set ‘pin_memory’, according 'dataset_class'
    if dataset_class == 'Visium16':
        pin_memory=False
    elif dataset_class == 'Visium64':
        pin_memory=False
    elif dataset_class == 'VisiumHD':
        pin_memory=False
    else:
        raise ValueError('Invalid dataset_class. Only "Visium16", "Visium64", and "VisiumHD" are supported.')

    ## Set up distributed sampler
    train_loader = DataLoader(train_dataset, batch_size, shuffle=True, num_workers=0, pin_memory=pin_memory, drop_last=True)
    test_loader = DataLoader(test_dataset, batch_size, shuffle=False, num_workers=0, pin_memory=pin_memory, drop_last=True)
    print("***** Finished building loaders *****")
    return train_loader, test_loader


def build_loaders_inference(batch_size, image_embed_path, spatial_pos_path, reduced_mtx_path, 
                            image_clacss='HIPT', dataset_class='Visium'):
    setup_seed(666)

    ## set image_paths, according 'image_clacss'
    if image_clacss == 'istar':
    
        dataset = DatasetCreat_istar(
            image_paths=image_embed_path,
            spatial_pos_path=spatial_pos_path,
            reduced_mtx_path=reduced_mtx_path,   
            dataset_class=dataset_class
        )
    elif image_clacss == 'HIPT' or image_clacss == 'Virchow2':
        import glob
        image_paths = glob.glob(str(image_embed_path))
        image_paths.sort()

        dataset = DatasetCreat(
            image_paths=image_paths,
            spatial_pos_path=spatial_pos_path,
            reduced_mtx_path=reduced_mtx_path,   
            dataset_class=dataset_class
        )
    else:
        raise ValueError('Invalid image_class. Only "istar" and "HIPT" are supported.')  
    
    ## set ‘pin_memory’, according 'dataset_class'
    if dataset_class == 'Visium16':
        pin_memory=False
    elif dataset_class == 'Visium64':
        pin_memory=False
    elif dataset_class == 'VisiumHD':
        pin_memory=False
    else:
        raise ValueError('Invalid dataset_class. Only "Visium16", "Visium64", and "VisiumHD" are supported.')

    all_dataset = DataLoader(dataset, batch_size, shuffle=False, num_workers=0, pin_memory=pin_memory, drop_last=False)
    print("***** Finished building loaders_inference *****")
    return all_dataset


def build_loaders_inference_allimage(batch_size, file_paths_spot, spatial_pos_path, 
                                     dataset_class='Visium16', file_paths_between_spot=None):
    setup_seed(666)

    if dataset_class == 'Visium16' or dataset_class == 'Visium64':
        if file_paths_between_spot is None:
            raise ValueError("file_paths_between_spot must be provided for Visium dataset")
        print("***** Building loaders_inference between spot *****")
        file_paths_spot = glob.glob(file_paths_spot)
        file_paths_between_spot = glob.glob(file_paths_between_spot)
        image_paths = file_paths_spot + file_paths_between_spot
        image_paths.sort()
    elif dataset_class == 'VisiumSC':
        print("***** Building loaders_inference sc image *****")
        image_paths = glob.glob(file_paths_spot)
        image_paths.sort()
    else:
        raise ValueError("Unknown dataset_class: {}".format(dataset_class))

    dataset = DatasetCreatImageBetweenSpot(
        image_paths=image_paths,
        spatial_pos_path=spatial_pos_path,
        dataset_class=dataset_class
    )
    
    all_dataset = DataLoader(dataset, batch_size=batch_size, shuffle=False, num_workers=0, pin_memory=False, drop_last=False)
    # all_dataset = DataLoader(dataset, batch_size=batch_size, shuffle=False, num_workers=0, pin_memory=True, drop_last=False)
    print("***** Finished building loaders_inference *****")
    return all_dataset



###############################################
# 2024.11.02 adjusted: add parameter： dataset
# 2025.01.08 design for istar image embeddings
###############################################
class DatasetCreat_istar(torch.utils.data.Dataset):
    def __init__(self, image_paths, spatial_pos_path, reduced_mtx_path, dataset_class):
        self.spatial_pos_csv = pd.read_csv(spatial_pos_path, sep=",", header=None)
        ## Load expression matrix:（Transport to: cell x features）
        ## if exits colnames and rownames，use 'allow_pickle=True'
        self.reduced_matrix = np.load(reduced_mtx_path).T   
        self.image_tensor = torch.load(image_paths)
        
        ## set ‘split_num’, according 'dataset_class'
        if dataset_class == 'Visium16':
            self.split_num = 16
        elif dataset_class == 'Visium64':
            self.split_num = 64
        elif dataset_class == 'VisiumHD':
            self.split_num = 4
        else:
            raise ValueError('Invalid dataset_class. Only "Visium" and "VisiumHD" are supported.')
        print("Finished loading all files")

    def __getitem__(self, idx):
        item = {}
        v1 = self.spatial_pos_csv.loc[idx, 0]   
        v2 = self.spatial_pos_csv.loc[idx, 1]  
        v3 = self.spatial_pos_csv.loc[idx, 2]   
        v4 = self.spatial_pos_csv.loc[idx, 3]  
    
        ## Stack the tensors in the list along a new dimension  
        item['image'] = self.image_tensor[idx * self.split_num : (idx + 1) * self.split_num]  
        item['reduced_expression'] = torch.tensor(self.reduced_matrix[idx, :]).float() 
        item['spatial_coords'] = [v1, v2]  
        item['array_row'] = v3   
        item['array_col'] = v4  
        return item

    def __len__(self):
        return len(self.reduced_matrix)
    

###############################################
# 2024.11.02 adjusted: add parameter： dataset
###############################################
class DatasetCreat(torch.utils.data.Dataset):
    def __init__(self, image_paths, spatial_pos_path, reduced_mtx_path, dataset_class):
        self.spatial_pos_csv = pd.read_csv(spatial_pos_path, sep=",", header=None)
        ## Load expression matrix:（Transport to: cell x features）
        ## if exits colnames and rownames，use 'allow_pickle=True'
        self.reduced_matrix = np.load(reduced_mtx_path).T    
        
        ## Load .pth files
        self.images = []
        for image_path in image_paths:
            if image_path.endswith('.pth'):
                image_tensor = torch.load(image_path)
                self.images.extend(image_tensor)
        self.image_data = torch.stack(self.images)
        self.image_tensor = self.image_data.view(self.image_data.size(0), -1)  
        
        ## set ‘split_num’, according 'dataset_class'
        if dataset_class == 'Visium16':
            self.split_num = 16
        elif dataset_class == 'Visium64':
            self.split_num = 64
        elif dataset_class == 'VisiumHD':
            self.split_num = 4
        else:
            raise ValueError('Invalid dataset_class. Only "Visium" and "VisiumHD" are supported.')                
        print("Finished loading all files")

    def __getitem__(self, idx):
        item = {}
        v1 = self.spatial_pos_csv.loc[idx, 0]   
        v2 = self.spatial_pos_csv.loc[idx, 1]  
        v3 = self.spatial_pos_csv.loc[idx, 2]   
        v4 = self.spatial_pos_csv.loc[idx, 3]  
    
        ## Stack the tensors in the list along a new dimension  
        item['image'] = self.image_tensor[idx * self.split_num : (idx + 1) * self.split_num]    
        item['reduced_expression'] = torch.tensor(self.reduced_matrix[idx, :]).float() 
        item['spatial_coords'] = [v1, v2]  
        item['array_row'] = v3   
        item['array_col'] = v4  
        return item

    def __len__(self):
        return len(self.reduced_matrix)
    

## 2023.10.26 need add CAE loss  
# class ContrastiveLoss(nn.Module):
#     def __init__(self, temperature=0.1):
#         super(ContrastiveLoss, self).__init__()
#         self.temperature = temperature

#     def forward(self, embeddings, labels):

#         batch_size = embeddings.shape[0]
        
#         embeddings = F.normalize(embeddings, p=2, dim=1)
#         sim_matrix = torch.exp(torch.matmul(embeddings, embeddings.T) / self.temperature)
       
#         pos_mask = (labels == 1).type(torch.bool)

#         neg_mask = ~pos_mask
#         neg_mask.fill_diagonal_(False)  # remove diagonal items

#         pos_similarities = (sim_matrix * pos_mask).sum(dim=1)
#         neg_similarities = (sim_matrix * neg_mask).sum(dim=1)
        
#         loss = -torch.mean(torch.log(pos_similarities / neg_similarities))

#         return loss


###############################
# CoSimLoss Loss： used
###############################
def CoSimLoss(Y_hat: torch.Tensor, Y: torch.Tensor) -> torch.Tensor:
    cos_by_col = nn.CosineSimilarity(dim=1)
    cos_by_row = nn.CosineSimilarity(dim=0)
    loss_col = (1 - cos_by_col(Y_hat, Y)).mean()
    loss_row = (1 - cos_by_row(Y_hat, Y)).mean()
    imp_loss = loss_col + loss_row
    return imp_loss

#######################################
# 2025.01.06  from Hist2ST
# ZINB_loss Loss： add
# https://github.com/biomed-AI/Hist2ST
#######################################
# def ZINBLoss(x, mean, disp, pi, scale_factor=1.0, ridge_lambda=0.0):
#     eps = 1e-10
#     if isinstance(scale_factor,float):
#         scale_factor=np.full((len(mean),),scale_factor)
#     scale_factor = scale_factor[:, None]
#     mean = mean * scale_factor

#     t1 = torch.lgamma(disp+eps) + torch.lgamma(x+1.0) - torch.lgamma(x+disp+eps)
#     t2 = (disp+x) * torch.log(1.0 + (mean/(disp+eps))) + (x * (torch.log(disp+eps) - torch.log(mean+eps)))
#     nb_final = t1 + t2

#     nb_case = nb_final - torch.log(1.0-pi+eps)
#     zero_nb = torch.pow(disp/(disp+mean+eps), disp)
#     zero_case = -torch.log(pi + ((1.0-pi)*zero_nb)+eps)
#     result = torch.where(torch.le(x, 1e-8), zero_case, nb_case)

#     if ridge_lambda > 0:
#         ridge = ridge_lambda*torch.square(pi)
#         result += ridge
#     result = torch.mean(result)
#     return result


#######################################
# 2025.01.06  from Hist2ST
# ZINB_loss Loss： add
# https://github.com/inoue0426/scVGAE/
#######################################
def ZINBLoss(y_true, y_pred_mean, y_pred_disp, y_pred_pi, eps=1e-10):
    """
    Compute the ZINB Loss.
        y_true: Ground truth data.
        y_pred_mean: \mu Predicted mean from the model.
        y_pred_disp: \theta Dispersion parameter.
        y_pred_pi: \pi Zero-inflation probability.
        eps: Small constant to prevent log(0).
    """

    ## Negative Binomial Loss
    nb_terms = (
        -torch.lgamma(y_true + y_pred_disp)
        + torch.lgamma(y_true + 1)
        + torch.lgamma(y_pred_disp)
        - y_pred_disp * torch.log(y_pred_disp + eps)
        + y_pred_disp * torch.log(y_pred_disp + y_pred_mean + eps)
        - y_true * torch.log(y_pred_mean + y_pred_disp + eps)
        + y_true * torch.log(y_pred_mean + eps)
    )

    ## Zero-Inflation
    zero_inflated = torch.log(y_pred_pi + (1 - y_pred_pi) * torch.pow(1 + y_pred_mean / y_pred_disp, -y_pred_disp))

    result = -torch.sum(
        torch.log(y_pred_pi + (1 - y_pred_pi) * torch.pow(1 + y_pred_mean / y_pred_disp, -y_pred_disp))
        * (y_true < eps).float()
        + (1 - (y_true < eps).float()) * nb_terms
    )

    return torch.round(result)
    

###############################
# PearsonCorrelationLoss： v1
###############################
def compute_pearson_corr(x, y, dim):
    x_mean = torch.mean(x, dim=dim, keepdim=True)
    y_mean = torch.mean(y, dim=dim, keepdim=True)
    x_std = torch.std(x, dim=dim, keepdim=True)
    y_std = torch.std(y, dim=dim, keepdim=True)

    ## Compute covariance of x and y along the specified dimension
    cov = torch.mean((x - x_mean) * (y - y_mean), dim=dim, keepdim=True)
    pearson_corr = cov / (x_std * y_std)
    loss = (1 - pearson_corr) / 2
    return torch.mean(loss)

def PearsonCorrelationLoss(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    ## Compute Pearson correlation loss along rows
    row_loss = compute_pearson_corr(x, y, dim=1)
    ## Compute Pearson correlation loss along columns
    col_loss = compute_pearson_corr(x, y, dim=0)
    combined_loss = row_loss + col_loss
    return combined_loss


#####################################################################
# ContrastiveLoss Loss: used
# 2025.01.03: Adjist weight parameter
#####################################################################
class ContrastiveLoss(nn.Module):
    def __init__(self, temperature=0.1, w1=1, w2=1, w3=1, w4=1, w5=1):
        super(ContrastiveLoss, self).__init__()
        self.temperature = temperature
        self.w1 = w1
        self.w2 = w2
        self.w3 = w3
        self.w4 = w4
        # self.w5 = w5

    def forward(
        self,
        embeddings_iamge,
        embeddings_matrix,
        labels,
        input_image_exp,
        reconstruction_iamge,
        reconstructed_matrix_all,
        reconstruction_iamge_reshapef2,
        input_matrix_exp,
        # recon_mat_mean, recon_mat_disp, recon_mat_pi
    ): 

        embeddings_iamge = F.normalize(embeddings_iamge, p=2, dim=1)
        embeddings_matrix = F.normalize(embeddings_matrix, p=2, dim=1)
        sim_matrix = torch.exp(torch.matmul(embeddings_iamge, embeddings_matrix.T) / self.temperature)
        
        pos_mask = (labels == 1).type(torch.bool)
        neg_mask = ~pos_mask

        pos_similarities = (sim_matrix * pos_mask).sum(dim=1)
        neg_similarities = (sim_matrix * neg_mask).sum(dim=1)

        #######################################################################
        ## 2023.12.20 adjust loss 

        # loss = (-self.w1*torch.mean(torch.log(pos_similarities / neg_similarities))            # contrast loss
        #         + self.w2*PearsonCorrelationLoss(reconstruction_iamge, input_image_exp)        # image loss
        #         + self.w3*PearsonCorrelationLoss(reconstructed_matrix_all, input_matrix_exp)   # matirx loss
        #         + self.w4*PearsonCorrelationLoss(reconstruction_iamge_reshapef2, input_matrix_exp))   # cross loss

        #######################################################################
        ## 2023.12.15 adjust loss 
        loss = (-self.w1*torch.mean(torch.log(pos_similarities / neg_similarities))     # contrast loss
                + self.w2*CoSimLoss(reconstruction_iamge, input_image_exp)              # image loss
                + self.w3*CoSimLoss(reconstructed_matrix_all, input_matrix_exp)         # matirx loss
                + self.w4*CoSimLoss(reconstruction_iamge_reshapef2, input_matrix_exp))  # cross loss
                # + self.w4*CoSimLoss(reconstruction_iamge_reshapef2, reconstructed_matrix_all))   # cross loss

        ##################################
        # 2025.01.06 add ZINB loss
        ##################################
        # - z_mean, z_dropout, z_dispersion: Outputs from the model, used for ZINB Loss calculation.
        # ZINBLoss(x_original, z_mean, z_dispersion, z_dropout)

        #######################################################################       
        ## 2023.12.15 adjust loss 
        # loss = (-self.w1*torch.mean(torch.log(pos_similarities / neg_similarities))       # contrast loss
        #         + self.w2*F.mse_loss(reconstruction_iamge, input_image_exp)               # image loss
        #         + self.w3*F.mse_loss(reconstructed_matrix_all, input_matrix_exp)          # matirx loss
        #         + self.w4*F.mse_loss(reconstruction_iamge_reshapef2, input_matrix_exp))   # cross loss
        #         # + self.w4*F.mse_loss(reconstruction_iamge_reshapef2, reconstructed_matrix_all)   # cross loss
        #         # + self.w5*ZINBLoss(input_matrix_exp, recon_mat_mean, recon_mat_disp, recon_mat_pi) ) # ZINB loss
        
        return loss
    

###################
# ProjectionHead  
###################
class ProjectionHead(nn.Module):
    def __init__(self, input_size, hidden_size, output_size):
        super(ProjectionHead, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden_size)   
        # self.gelu = nn.GELU()  
        self.relu = nn.ReLU(inplace=True)  
        self.fc2 = nn.Linear(hidden_size, output_size)   

    def forward(self, x):
        x = self.fc1(x)  
        # x = self.gelu(x)   # BLEEP
        x = self.relu(x)  
        x = self.fc2(x)   
        return x
      
class ELU(nn.Module):
    def __init__(self, alpha, beta):
        super().__init__()
        self.activation = nn.ELU(alpha=alpha, inplace=True)
        self.beta = beta
    def forward(self, x):
        return self.activation(x) + self.beta

##############################
# Spatial data embedding  
##############################
class Autoencoder_matrix(nn.Module):    # 2 layers
    def __init__(self, input_dim, latent_dim, n_encoder_hidden_matrix, negative_slope=0.01):
        super(Autoencoder_matrix, self).__init__()
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, n_encoder_hidden_matrix),
            nn.LeakyReLU(negative_slope),    # GELU()  BLEEP
            nn.Linear(n_encoder_hidden_matrix, latent_dim)
        )
        self.decoder = nn.Sequential(
            nn.Linear(latent_dim, n_encoder_hidden_matrix),
            nn.LeakyReLU(negative_slope),
            nn.Linear(n_encoder_hidden_matrix, input_dim),
            # nn.Sigmoid(),
            # nn.Softplus(),     # xfuse   positive
            ELU(alpha=0.01, beta=0.01),    # iSTAR     positive
        )
    def forward(self, x):
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return decoded


##############################
# 2025.01.06 Adjust, add ZINB
# Spatial data embedding  
##############################
# class Autoencoder_matrix(nn.Module):    # 2 layers
#     def __init__(self, input_dim, latent_dim, n_encoder_hidden_matrix, negative_slope=0.01):
#         super(Autoencoder_matrix, self).__init__()
#         self.encoder = nn.Sequential(
#             nn.Linear(input_dim, n_encoder_hidden_matrix),
#             nn.LeakyReLU(negative_slope),    # GELU()  BLEEP
#             nn.Linear(n_encoder_hidden_matrix, latent_dim)
#         )
#         self.decoder = nn.Sequential(
#             nn.Linear(latent_dim, n_encoder_hidden_matrix),
#             nn.LeakyReLU(negative_slope),
#             nn.Linear(n_encoder_hidden_matrix, input_dim),
#             # nn.Sigmoid(),
#             # nn.Softplus(),     # xfuse   positive
#             ELU(alpha=0.01, beta=0.01),    # iSTAR     positive
#         )

#         self.decoder_mean = nn.Sequential(
#             nn.Linear(latent_dim, n_encoder_hidden_matrix),
#             nn.LeakyReLU(negative_slope),
#             nn.Linear(n_encoder_hidden_matrix, input_dim),
#             nn.Softplus()
#             # ELU(alpha=0.01, beta=0.01),    # iSTAR     positive
#         )
#         self.decoder_disp = nn.Sequential(
#             nn.Linear(latent_dim, n_encoder_hidden_matrix),
#             nn.LeakyReLU(negative_slope),
#             nn.Linear(n_encoder_hidden_matrix, input_dim),
#             nn.Softplus()
#             # ELU(alpha=0.01, beta=0.01),    # iSTAR     positive
#         )
#         self.decoder_pi = nn.Sequential(
#             nn.Linear(latent_dim, n_encoder_hidden_matrix),
#             nn.LeakyReLU(negative_slope),
#             nn.Linear(n_encoder_hidden_matrix, input_dim),
#             nn.Sigmoid()
#         )

#     def forward(self, x):
#         encoded = self.encoder(x)
#         decoded = self.decoder(encoded)

#         decoded_mean = self.decoder_mean(encoded)
#         decoded_disp = self.decoder_disp(encoded)
#         decoded_pi = self.decoder_pi(encoded)

#         return decoded, decoded_mean, decoded_disp, decoded_pi
    

###############################
# Patch image embedding  
###############################
class Autoencoder_image(nn.Module):
    def __init__(self, input_dim, latent_dim, n_encoder_hidden_image, negative_slope=0.01):
        super(Autoencoder_image, self).__init__()
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, n_encoder_hidden_image),
            nn.LeakyReLU(negative_slope),
            nn.Linear(n_encoder_hidden_image, latent_dim),
        )
        self.decoder = nn.Sequential(
            nn.Linear(latent_dim, n_encoder_hidden_image),
            nn.LeakyReLU(negative_slope),
            nn.Linear(n_encoder_hidden_image, input_dim),
            # nn.Softplus(),     # xfuse   positivate
            ELU(alpha=0.01, beta=0.01),    # iSTAR     positivate
        )
    def forward(self, x):
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return decoded
    

########################
# FineSTModel    
########################
class FineSTModel(nn.Module):
    def __init__(
        self, 
        n_input_matrix=578, 
        n_input_image=384, 
        n_encoder_hidden_matrix=512,
        n_encoder_hidden_image=256,
        n_encoder_hidden=512,  # not use
        n_encoder_latent=128, 
        n_projection_hidden=256, 
        n_projection_output=128, 
        n_encoder_layers=2, 
        dropout_rate=0, 
        negative_slope=0.01
    ):    
        super(FineSTModel,self).__init__()     
        self.Autoencoder_matrix = Autoencoder_matrix(n_input_matrix, n_encoder_latent, n_encoder_hidden_matrix) 
        self.matrix_encoder = self.Autoencoder_matrix.encoder
        self.matrix_decoder = self.Autoencoder_matrix.decoder
        self.Autoencoder_image = Autoencoder_image(n_input_image, n_encoder_latent, n_encoder_hidden_image)  
        self.image_encoder = self.Autoencoder_image.encoder
        self.image_decoder = self.Autoencoder_image.decoder
        self.image_projection = ProjectionHead(n_encoder_latent, n_projection_hidden, n_projection_output)
        self.matrix_projection = ProjectionHead(n_encoder_latent, n_projection_hidden, n_projection_output)
                        
    def forward(self, x, y):              
        representation_matrix = self.matrix_encoder(x) 
        reconstruction_matrix = self.matrix_decoder(representation_matrix) 
        projection_matrix = self.matrix_projection(representation_matrix)       
        representation_image = self.image_encoder(y) 
        reconstruction_image = self.image_decoder(representation_image) 
        projection_image = self.image_projection(representation_image)      
        return (representation_matrix, reconstruction_matrix, projection_matrix, 
                representation_image, reconstruction_image, projection_image)


#################################
# 2025.01.06 adjust, add ZINB
# FineSTModel   
#################################
# class FineSTModel(nn.Module):
#     def __init__(
#         self, 
#         n_input_matrix=578, 
#         n_input_image=384, 
#         n_encoder_hidden_matrix=512,
#         n_encoder_hidden_image=256,
#         n_encoder_hidden=512,  # not use
#         n_encoder_latent=128, 
#         n_projection_hidden=256, 
#         n_projection_output=128, 
#         n_encoder_layers=2, 
#         dropout_rate=0, 
#         negative_slope=0.01
#     ):
#         super(FineSTModel,self).__init__()     
#         self.Autoencoder_matrix = Autoencoder_matrix(n_input_matrix, n_encoder_latent, n_encoder_hidden_matrix) 
#         self.matrix_encoder = self.Autoencoder_matrix.encoder
#         self.matrix_decoder = self.Autoencoder_matrix.decoder
#         self.matrix_decoder_mean = self.Autoencoder_matrix.decoder_mean
#         self.matrix_decoder_disp = self.Autoencoder_matrix.decoder_disp
#         self.matrix_decoder_pi = self.Autoencoder_matrix.decoder_pi
#         self.Autoencoder_image = Autoencoder_image(n_input_image, n_encoder_latent, n_encoder_hidden_image)  
#         self.image_encoder = self.Autoencoder_image.encoder
#         self.image_decoder = self.Autoencoder_image.decoder
#         self.image_projection = ProjectionHead(n_encoder_latent, n_projection_hidden, n_projection_output)
#         self.matrix_projection = ProjectionHead(n_encoder_latent, n_projection_hidden, n_projection_output)

#     def forward(self, x, y):              
#         representation_matrix = self.matrix_encoder(x) 
#         reconstruction_matrix = self.matrix_decoder(representation_matrix) 

#         recon_mat_mean = self.matrix_decoder_mean(representation_matrix) 
#         recon_mat_disp = self.matrix_decoder_disp(representation_matrix) 
#         recon_mat_pi = self.matrix_decoder_pi(representation_matrix) 

#         projection_matrix = self.matrix_projection(representation_matrix)       
#         representation_image = self.image_encoder(y) 
#         reconstruction_image = self.image_decoder(representation_image) 
#         projection_image = self.image_projection(representation_image) 

#         return (representation_matrix, 
#                 reconstruction_matrix, 
#                 projection_matrix, 
#                 representation_image, 
#                 reconstruction_image, 
#                 projection_image,
#                 recon_mat_mean, recon_mat_disp, recon_mat_pi)


def load_model(dir_name, parameter_file_path, params, gene_hv):    
    save_folder = os.path.join(dir_name, "epoch_"+str(params["training_epoch"])+".pt")
    if(dev=="cpu"):
        checkpoint = torch.load(save_folder,map_location="cpu")
    else:
        checkpoint = torch.load(save_folder)
    
    model_state_dict = checkpoint['model_state_dict']
    
    ## load parameter settings
    with open(parameter_file_path,"r") as json_file:
        params = json.load(json_file)
    params['n_input_matrix'] = len(gene_hv)
    # params['n_input_image'] = 384
    
    ## init the model
    model = FineSTModel(n_input_matrix=params['n_input_matrix'],
                              n_input_image=params['n_input_image'],
                              n_encoder_hidden_matrix=params["n_encoder_hidden_matrix"],
                              n_encoder_hidden_image=params["n_encoder_hidden_image"],
                              n_encoder_latent=params["n_encoder_latent"],
                              n_projection_hidden=params["n_projection_hidden"],
                              n_projection_output=params["n_projection_output"],
                              n_encoder_layers=params["n_encoder_layers"]).to(device)    
    ## load model states
    model.load_state_dict(model_state_dict)    
    return model



import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F

class MeanAct(nn.Module):
    def __init__(self):
        super(MeanAct, self).__init__()
    def forward(self, x):
        return torch.clamp(torch.exp(x), min=1e-5, max=1e6)

class DispAct(nn.Module):
    def __init__(self):
        super(DispAct, self).__init__()
    def forward(self, x):
        return torch.clamp(F.softplus(x), min=1e-4, max=1e4)
    
def NB_loss(x, h_r, h_p):
        
        ll = torch.lgamma(torch.exp(h_r) + x) - torch.lgamma(torch.exp(h_r))
        ll += h_p * x - torch.log(torch.exp(h_p) + 1) * (x + torch.exp(h_r))

        loss = -torch.mean(torch.sum(ll, axis=-1))
        return loss

def ZINB_loss(x, mean, disp, pi, scale_factor=1.0, ridge_lambda=0.0):
    eps = 1e-10
    if isinstance(scale_factor,float):
        scale_factor=np.full((len(mean),),scale_factor)
    scale_factor = scale_factor[:, None]
    mean = mean * scale_factor

    t1 = torch.lgamma(disp+eps) + torch.lgamma(x+1.0) - torch.lgamma(x+disp+eps)
    t2 = (disp+x) * torch.log(1.0 + (mean/(disp+eps))) + (x * (torch.log(disp+eps) - torch.log(mean+eps)))
    nb_final = t1 + t2

    nb_case = nb_final - torch.log(1.0-pi+eps)
    zero_nb = torch.pow(disp/(disp+mean+eps), disp)
    zero_case = -torch.log(pi + ((1.0-pi)*zero_nb)+eps)
    result = torch.where(torch.le(x, 1e-8), zero_case, nb_case)

    if ridge_lambda > 0:
        ridge = ridge_lambda*torch.square(pi)
        result += ridge
    result = torch.mean(result)
    return result


import logging
logging.getLogger().setLevel(logging.INFO)
from .utils import *
from .loadData import *
import numpy as np
from scipy.spatial import cKDTree


def perform_inference_image(model, test_loader, dataset_class='Visium'):
    print("device",device)    

    #####################################################################################
    # for whole dataset
    #####################################################################################        
    print("***** Begin perform_inference: ******")
    
    input_spot_all, input_image_all, input_coord_all, _, _ = extract_test_data(test_loader)
            
    ## input image and matrix
    matrix_profile = input_spot_all.to(device)
    image_profile = input_image_all.to(device)
    ## reshape image
    image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [1331, 256, 384] --> [1331*256, 384]
    input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
    
    ## useful model
    representation_matrix = model.matrix_encoder(matrix_profile)
    reconstructed_matrix = model.matrix_decoder(representation_matrix)
    projection_matrix = model.matrix_projection(representation_matrix)  
    representation_image = model.image_encoder(input_image_exp) 
    reconstruction_iamge = model.image_decoder(representation_image)
    projection_image = model.image_projection(representation_image)

    ## reshape
    _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
    _, projection_image_reshape = reshape_latent_image(projection_image, dataset_class)

    ## cross decoder
    reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
    _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)
    
    
    #####################################################################################  
    # convert
    #####################################################################################  
    ## matrix
    matrix_profile = matrix_profile.cpu().detach().numpy() 
    reconstructed_matrix = reconstructed_matrix.cpu().detach().numpy() 
    reconstruction_iamge_reshapef2 = reconstruction_iamge_reshapef2.cpu().detach().numpy() 
    ## latent space
    representation_image_reshape = representation_image_reshape.cpu().detach().numpy() 
    representation_matrix = representation_matrix.cpu().detach().numpy() 
    ## latent space -- peojection
    projection_image_reshape = projection_image_reshape.cpu().detach().numpy() 
    projection_matrix = projection_matrix.cpu().detach().numpy() 
    ## image
    input_image_exp = input_image_exp.cpu().detach().numpy() 
    reconstruction_iamge = reconstruction_iamge.cpu().detach().numpy()
    # ## tensor recon_f2
    # reconstructed_matrix_reshaped = reconstructed_matrix_reshaped.cpu().detach().numpy()
    
    return (matrix_profile, 
            reconstructed_matrix, 
            reconstruction_iamge_reshapef2, 
            representation_image_reshape,
            representation_matrix,
            projection_image_reshape,
            projection_matrix,
            input_image_exp,
            reconstruction_iamge,
            reconstructed_matrix_reshaped,
            input_coord_all)



def perform_inference_image_between_spot(model, test_loader, dataset_class='Visium'):
    print("device",device)    

    #####################################################################################
    # for whole dataset
    #####################################################################################        
    print("***** Begin perform_inference: ******")
    
    input_image_all, input_coord_all = extract_test_data_image_between_spot(test_loader)   
            
    ## input image
    image_profile = input_image_all.to(device)
    ## reshape image
    image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [adata.shape[0], 256, 384] --> [adata.shape[0]*256, 384]
    input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
    ## useful model
    representation_image = model.image_encoder(input_image_exp) 
    ## cross decoder
    reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
    _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)
    ## reshape
    _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
    
    #####################################################################################  
    # convert
    #####################################################################################  
    ## matrix
    representation_image_reshape = representation_image_reshape.cpu().detach().numpy() 
    reconstruction_iamge_reshapef2 = reconstruction_iamge_reshapef2.cpu().detach().numpy() 
    
    return (reconstruction_iamge_reshapef2, 
            reconstructed_matrix_reshaped,
            representation_image_reshape,
            input_image_exp,
            input_coord_all)

##################################################################################
# imputation.py
##################################################################################

## Find the nearest point in adata_know for each point in adata_spot
def find_nearest_point(adata_spot, adata_know):
    nearest_points = []
    for point in adata_spot:
        distances = np.linalg.norm(adata_know - point, axis=1)
        nearest_index = np.argmin(distances)
        nearest_points.append(adata_know[nearest_index])
    return np.array(nearest_points)

##################################################################################
# using 7 neighborhood： cKDTree is more faster
# Find k nearest neighbors for each point in nearest_points within adata_know
##################################################################################

## Function 1: Using cKDTree
def find_nearest_neighbors(nearest_points, adata_know, k=6):
    nbs = []
    nbs_indices = []
    tree = cKDTree(adata_know)
    for point in nearest_points:
        dist, indices = tree.query(point, k+1)
        nbs.append(adata_know[indices])
        nbs_indices.append(indices)
    return np.array(nbs), np.array(nbs_indices)

# ## Function 2: Using NearestNeighbors
# from sklearn.neighbors import NearestNeighbors
# def find_nearest_neighbors(nearest_points, adata_know, k=6):
#     nbs = []
#     nbs_indices = []
#     nbrs = NearestNeighbors(n_neighbors=k+1, algorithm='ball_tree', metric='euclidean').fit(adata_know)
#     for point in nearest_points:
#         distances, indices = nbrs.kneighbors([point])
#         nbs.append(adata_know[indices][0])
#         nbs_indices.append(indices[0])
#     return np.array(nbs), np.array(nbs_indices)

## Calculate Euclidean distances between each point in adata_spot and its nearest neighbors
def calculate_euclidean_distances(adata_spot, nbs):
    distances = []
    for point, neighbors in zip(adata_spot, nbs):
        dist = np.linalg.norm(neighbors - point, axis=1)
        distances.append(dist)
    return np.array(distances)




import pandas as pd
import numpy as np
from sklearn.neighbors import NearestNeighbors
from statsmodels.stats.multitest import fdrcorrection
from threadpoolctl import threadpool_limits
from .utils import *
from itertools import zip_longest
import anndata as ann
import scipy
import pyreadr
import requests





#######################################
# 2024.11.28 Add code of LR-TF
# , datahost='builtin' -- !!
#######################################
def extract_tf(species, datahost='package'):
    """
    find overlapping LRs from CellChatDB
    :param species: support 'human', 'mouse' and 'zebrafish'
    :param datahost: the host of the ligand-receptor data. 
                    'builtin' for package built-in otherwise from figshare
    :return: LR_TF (containing comprehensive info from CellChatDB) dataframe
    """
    
    if datahost == 'package':

        if species in ['mouse', 'human']:
            # datapath = '/mnt/lingyu/nfs_share2/Python/FineST/FineST/FineST/datasets/TF_data/%s-' %(species)
            # datapath = './datasets/TF_data/%s-' %(species)
            datapath = './FineST/datasets/TF_data/%s-' %(species)
        else:
            raise ValueError("species type: {} is not supported currently. Please have a check.".format(species))
        
        LR_TF = pyreadr.read_r(datapath + 'TF_PPRhuman.rda')['TF_PPRhuman']

    else:
        if species == 'mouse':
            url = 'https://figshare.com/ndownloader/files/50860644'
        elif species == 'human':
            url = 'https://figshare.com/ndownloader/files/50860650'
        else:
            raise ValueError("species type: {} is not supported currently. Please have a check.".format(species))
        
        # specify where to download the file
        # download_path = '/mnt/lingyu/nfs_share2/Python/FineST/FineST/FineST/datasets/TF_data/temp.rda'
        # download_path = './datasets/TF_data/'
        download_path = './FineST/datasets/TF_data/'
        
        # download the file
        r = requests.get(url)
        with open(download_path, 'wb') as f:
            f.write(r.content)

        LR_TF = pyreadr.read_r(download_path)['TF_PPRhuman']

        # remove the downloaded file after use
        os.remove(download_path)

        # if species == 'mouse':
        #     LR_TF = pyreadr.read_r('https://figshare.com/ndownloader/files/50860644')['TF_PPRhuman']
        #     # TF_TG = pyreadr.read_r('https://figshare.com/ndownloader/files/50860656')
        # elif species == 'human':
        #     LR_TF = pyreadr.read_r('https://figshare.com/ndownloader/files/50860650')['TF_PPRhuman']
        #     # TF_TG = pyreadr.read_r('https://figshare.com/ndownloader/files/50860647')
        # else:
        #     raise ValueError("species type: {} is not supported currently. Please have a check.".format(species))
        
    return LR_TF


def top_pattern_LR2TF(tmp, ligand_list, receptor_list, top_num=20):
    """
    The function takes in a DataFrame and two lists of ligands and receptors respectively, 
    filter the DataFrame based on the lists and return the top rows sorted by 'value' column.
    Args:
    tmp : a DataFrame to process
    ligand_list : a list of ligands to filter on
    receptor_list : a list of receptors to filter on
    top_num : the number of top rows to return, defaults to 20

    Returns:
    tmp_df : a DataFrame after processing
    """

    #################################################
    # Ligand or Receptor contain pattern LR external
    #################################################
    # # Filter the DataFrame based on the ligand_list and receptor_list
    # lData = [tmp[tmp['Ligand'].str.contains(ligand, na=False)] for ligand in ligand_list]
    # rData = [tmp[tmp['Receptor'].str.contains(receptor, na=False)] for receptor in receptor_list]
    # print("Ligand in R2TFdatabase:", len(lData))
    # print("Receptor in R2TFdatabase:", len(rData))
    
    # # Concatenate the filtered DataFrame and drop duplicates
    # fData = pd.concat(lData + rData).drop_duplicates()
    # print("Ligand or Receptor in R2TFdatabase:", fData.shape[0])
    
    #################################################
    # Ligand or Receptor only contain pattern LR 
    #################################################
    # Filter the DataFrame based on the ligand_list and receptor_list
    fData = tmp[tmp['Ligand'].isin(ligand_list) & tmp['Receptor'].isin(receptor_list)]
    print("Ligand and Receptor in R2TFdatabase:", fData.shape[0])

    # Sort the DataFrame by 'value' column and return the top rows
    tmp_df = fData.sort_values(by='value', ascending=False).head(top_num)

    # rename colname
    subdf = pd.DataFrame(tmp_df.rename(columns={"Ligand": "Ligand_symbol", 
                                                "Receptor": "Receptor_symbol", 
                                                "tf": "TF", "value": "value"}) )    
    
    return subdf


def pattern_LR2TF2TG(histology_results, pattern_num, R_TFdatabase, TF_TGdatabase):
    """
    The function takes in a DataFrame, checks if column 'g' contains two '_', if so, it splits the 'g' column value into two new rows
    Args:
    histology_results : a DataFrame to process
    pattern_num : the pattern number to filter on, defaults to 0
    R_TFdatabase : the DataFrame containing receptor to TF mapping

    Returns:
    tmp : a DataFrame after processing
    """
    rows = []

    for i, row in histology_results.iterrows():
        # check if column 'g' contains two '_'
        if row['g'].count('_') == 2:
            # split the 'g' column value into three parts
            gene1, gene2, gene3 = row['g'].split('_')

            # create two new rows, gene1_gene2 and gene1_gene3 respectively
            new_row1 = row.copy()
            new_row1['g'] = gene1 + '_' + gene2
            new_row2 = row.copy()
            new_row2['g'] = gene1 + '_' + gene3

            # add the new rows to the result DataFrame
            rows.append(new_row1)
            rows.append(new_row2)
        else:
            # if column 'g' does not contain two '_', add the original row to the result DataFrame directly
            rows.append(row)

    # after the loop, create DataFrame at once
    p0_results = pd.DataFrame(rows, columns=histology_results.columns)
    
    LRp0 = p0_results[p0_results['pattern']==pattern_num]['g']

    Lp0 = [gene for pair in LRp0 for gene in pair.split('_')[0:1]]
    print("This pattern contain %s unique ligand", len(set(Lp0)))

    Rp0 = [gene for pair in LRp0 for gene in pair.split('_')[1:]]
    print("This pattern contain %s unique receptor", len(set(Rp0)))

    R_TFdata_df = R_TFdatabase[R_TFdatabase['receptor'].isin(Rp0)]
    ligand = Lp0
    receptor = Rp0

    result = pd.concat([pd.DataFrame(ligand), pd.DataFrame(receptor)], axis=1)
    result.columns = ['ligand', 'receptor']
    result = result.dropna()

    comm = result.merge(R_TFdata_df, on='receptor', how='left')
    comm = comm.dropna()

    tf_comm = [gene for gene in comm['tf']]
    print("This pattern contain %s unique tf", len(set(tf_comm)))
    
    R_TFdata_TG_df = TF_TGdatabase[TF_TGdatabase['tf'].isin(tf_comm)]
    R_TFdata_TG_df
    comm_all = comm.merge(R_TFdata_TG_df, on='tf', how='right')
    comm_all = comm_all.dropna().drop_duplicates()
    comm_all
    
    tmp = comm_all[["ligand","receptor","tf", "target", "tf_PPR"]]
    tmp = tmp.rename(columns={"ligand": "Ligand", "receptor": "Receptor", "tf": "tf", "target": "Target", "tf_PPR": "value"})
    tmp = tmp.drop_duplicates()
    
    return tmp


def pattern_LR2TF(histology_results, pattern_num, R_TFdatabase):
    """
    The function takes in a DataFrame, checks if column 'g' contains two '_', if so, it splits the 'g' column value into two new rows
    Args:
    histology_results : a DataFrame to process
    pattern_num : the pattern number to filter on, defaults to 0
    R_TFdatabase : the DataFrame containing receptor to TF mapping

    Returns:
    tmp : a DataFrame after processing
    """
    rows = []

    for i, row in histology_results.iterrows():
        # check if column 'g' contains two '_'
        if row['g'].count('_') == 2:
            # split the 'g' column value into three parts
            gene1, gene2, gene3 = row['g'].split('_')

            # create two new rows, gene1_gene2 and gene1_gene3 respectively
            new_row1 = row.copy()
            new_row1['g'] = gene1 + '_' + gene2
            new_row2 = row.copy()
            new_row2['g'] = gene1 + '_' + gene3

            # add the new rows to the result DataFrame
            rows.append(new_row1)
            rows.append(new_row2)
        else:
            # if column 'g' does not contain two '_', add the original row to the result DataFrame directly
            rows.append(row)

    # after the loop, create DataFrame at once
    p0_results = pd.DataFrame(rows, columns=histology_results.columns)
    
    LRp0 = p0_results[p0_results['pattern']==pattern_num]['g']

    Lp0 = [gene for pair in LRp0 for gene in pair.split('_')[0:1]]
    print("This pattern contain %s unique ligand", len(set(Lp0)))

    Rp0 = [gene for pair in LRp0 for gene in pair.split('_')[1:]]
    print("This pattern contain %s unique receptor", len(set(Rp0)))

    R_TFdata_df = R_TFdatabase[R_TFdatabase['receptor'].isin(Rp0)]
    ligand = Lp0
    receptor = Rp0

    result = pd.concat([pd.DataFrame(ligand), pd.DataFrame(receptor)], axis=1)
    result.columns = ['ligand', 'receptor']
    result = result.dropna()

    comm = result.merge(R_TFdata_df, on='receptor', how='left')
    comm = comm.dropna()

    tmp = comm[["ligand","receptor","tf","tf_PPR"]]
    tmp = tmp.rename(columns={"ligand": "Ligand", "receptor": "Receptor", "tf": "tf", "tf_PPR": "value"})
    tmp = tmp.drop_duplicates()
    
    return tmp


#######################################
# 2024.11.17 Add code of SpatialDM
#######################################
def LRpair_gene(df):
    """
    df : DataFrame

    ## see the unique gene of sig LR pairs

    Returns a DataFrame of unique elements from  'Ligand0', 'Ligand1', 'Receptor0', 
    'Receptor1', and 'Receptor2' columns of the DataFrame where 'selected' is True.
    """
    # Filter the DataFrame
    filtered_df = df[df['selected'] == True]

    # Get the unique elements
    unique_elements = set(filtered_df['Ligand0'].tolist() + 
                          filtered_df['Ligand1'].tolist() + 
                          filtered_df['Receptor0'].tolist() + 
                          filtered_df['Receptor1'].tolist() + 
                          filtered_df['Receptor2'].tolist())
    unique_elements_df = pd.DataFrame(unique_elements)

    return unique_elements_df


def anno_LRpair(adata_impt_all):
    """
    adata_impt_all : AnnData object

    Returns a DataFrame resulting from merging 
    adata_impt_all.uns['global_res'].sort_values(by='fdr')  with 
    adata_impt_all.uns['geneInter'] on 'Ligand0' and 'interaction_name'.
    """

    # Create a DataFrame from 'geneInter'
    geneInter_df = adata_impt_all.uns['geneInter']
    spa_coexp_pair = adata_impt_all.uns['global_res'].sort_values(by='fdr')  

    # Merge the dataframes
    merged_df = pd.merge(spa_coexp_pair, geneInter_df, how='left', 
                         left_index=True, right_index=True)

    # Keep only the columns of interest
    final_df = merged_df[['Ligand0', 'Ligand1', 'Receptor0', 'Receptor1', 
                          'Receptor2', 'z_pval', 'z', 'fdr', 'selected', 
                          'evidence', 'annotation']]

    return final_df


#######################################
# 2024.11.11.Adjust code of SpatialDM
#######################################

def _Euclidean_to_RBF(X, l, singlecell):
    """Convert Euclidean distance to RBF distance"""
    from scipy.sparse import issparse
    if issparse:
        rbf_d = X
        rbf_d[X.nonzero()] = np.exp(-X[X.nonzero()].A**2 / (2 * l ** 2))
    else:
        rbf_d = np.exp(- X**2 / (2 * l ** 2))
    
    # At single-cell resolution, no within-spot communications
    if singlecell:

        ###################
        # old for SpatialDM
        ###################
        # np.fill_diagonal(rbf_d, 0)

        rbf_d_dense = rbf_d.toarray()  # or rbf_d.todense()
        np.fill_diagonal(rbf_d_dense, 0)

    else:
        rbf_d.setdiag(np.exp(-X.diagonal()**2 / (2 * l ** 2)))

    return rbf_d


#######################################
# 2024.11.11.Adjust code of SpatialDM
#######################################
def weight_matrix(adata_impt_all, l, cutoff, single_cell=False, n_nearest_neighbors=6):

    """
    Compute weight matrix based on radial basis kernel, more efficient than SpatialDM.
    cutoff & n_neighbors are two alternative options to restrict signaling range.
    :param l: radial basis kernel parameter, need to be customized for optimal weight gradient and \
    to restrain the range of signaling before downstream processing.
    :param cutoff: (for secreted signaling) minimum weight to be kept from the rbf weight matrix. \
    Weight below cutoff will be made zero
    :param n_neighbors: (for secreted signaling) number of neighbors per spot from the rbf weight matrix.
    :param n_nearest_neighbors: (for adjacent signaling) number of neighbors per spot from the rbf \
    weight matrix.
    Non-neighbors will be made 0
    :param single_cell: if single cell resolution, diagonal will be made 0.
    :return: secreted signaling weight matrix: adata.obsp['weight'], \
            and adjacent signaling weight matrix: adata.obsp['nearest_neighbors']
    """

    adata_impt_all.uns['single_cell'] = single_cell
    if isinstance(adata_impt_all.obsm['spatial'], pd.DataFrame):
        X_loc = adata_impt_all.obsm['spatial'].values
    else:
        X_loc = adata_impt_all.obsm['spatial']

    n_neighbors = n_nearest_neighbors * 31

    #####################################
    # large neighborhood for W (5 layers)
    #####################################
    nnbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm='ball_tree', metric='euclidean').fit(X_loc)
    nbr_d = nnbrs.kneighbors_graph(X_loc, mode='distance')
    rbf_d = _Euclidean_to_RBF(nbr_d, l, single_cell)

    #####################################
    # small neighborhood for RBF
    #####################################
    nnbrs0 = NearestNeighbors(n_neighbors=n_nearest_neighbors, 
                              algorithm='ball_tree', metric='euclidean').fit(X_loc)
    nbr_d0 = nnbrs0.kneighbors_graph(X_loc, mode='distance')
    rbf_d0 = _Euclidean_to_RBF(nbr_d0, l, single_cell)

    #########################################################################
    # NOTE: add more info about cutoff, n_neighbors and n_nearest_neighbors
    #########################################################################
    if cutoff:
        #####################################
        # not efficient -- old code
        #####################################
        # rbf_d[rbf_d < cutoff] = 0
        
        #####################################################################
        # more efficient -- new code
        # https://seanlaw.github.io/2019/02/27/set-values-in-sparse-matrix/
        #####################################################################
        nonzero_mask = np.array(rbf_d[rbf_d.nonzero()] < cutoff)[0]
        rows = rbf_d.nonzero()[0][nonzero_mask]
        cols = rbf_d.nonzero()[1][nonzero_mask]
        rbf_d[rows, cols] = 0
    
    # elif n_neighbors:
    #     nbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm='ball_tree').fit(rbf_d)
    #     knn = nbrs.kneighbors_graph(rbf_d).toarray()
    #     rbf_d = rbf_d * knn

    adata_impt_all.obsp['weight'] = rbf_d * adata_impt_all.shape[0] / rbf_d.sum()
    adata_impt_all.obsp['nearest_neighbors'] = rbf_d0 * adata_impt_all.shape[0] / rbf_d0.sum()

    return adata_impt_all


#######################################
# 2024.11.11.Adjust code of SpatialDM
#######################################
def pathway_analysis(sample=None,
                    all_interactions=None,
                    groups=None, cut_off=None,  # LLY add
        interaction_ls=None, name=None, dic=None):
    """
    Compute enriched pathways for a list of pairs or a dic of SpatialDE results.
    :param sample: spatialdm obj
    :param ls: a list of LR interaction names for the enrichment analysis
    :param path_name: str. For later recall sample.path_summary[path_name]
    :param dic: a dic of SpatialDE results (See tutorial)
    """
    if interaction_ls is not None:
        dic = {name: interaction_ls}
    if sample is not None:
        ## the old one
        all_interactions = sample.uns['geneInter']
        ## Load the original object from a pickle file when needed
        # with open(sample.uns["geneInter"], "rb") as f:
        #     all_interactions = pickle.load(f)

    # print(all_interactions)
    df = pd.DataFrame(all_interactions.groupby('pathway_name').interaction_name)
    df = df.set_index(0)
    total_feature_num = len(all_interactions)
    result = []
    result_pattern = []
    for n,ls in dic.items():
        qset = set([x.upper() for x in ls]).intersection(all_interactions.index)
        query_set_size = len(qset)
        for modulename, members in df.iterrows():
            module_size = len(members.values[0])
            overlap_features = qset.intersection(members.values[0])
            overlap_size = len(overlap_features)
    
            negneg = total_feature_num + overlap_size - module_size - query_set_size
            # Fisher's exact test
            p_FET = scipy.stats.fisher_exact([[overlap_size, query_set_size - overlap_size],
                                        [module_size - overlap_size, negneg]], 'greater')[1]

            result.append((p_FET, modulename, module_size, overlap_size, overlap_features, n))
            result_pattern.append((p_FET, modulename, module_size, overlap_size, overlap_features, n, query_set_size, negneg))  # 将结果添加到列表中

    result = pd.DataFrame(result).set_index(1)
    result.columns = ['fisher_p', 'pathway_size', 'selected', 'selected_inters', 'name']

    result_pattern = pd.DataFrame(result_pattern).set_index(1)  # 将结果转换为数据框并设置索引为第二列
    result_pattern.columns = ['fisher_p', 'module_size', 'overlap_size', 'selected_inters', 'name', 'query_set_size', 'negneg']  # 设置数据框的列名
    
    if sample is not None:
        sample.uns['pathway_summary'] = result    

    #################################################################################
    # copy form Function dot_path(), get all pathway information
    #################################################################################
    pathway_res = result[result.selected >= cut_off]
    if groups is not None:
        pathway_res = pathway_res.loc[pathway_res.name.isin(groups)]

    #################################################################################
    # copy form Function dot(), get pattern visulization: selected pathway 
    #################################################################################
    for i, name in enumerate(pathway_res.name.unique()):
        result1 = pathway_res.loc[pathway_res.name == name]
        result1 = result1.sort_values('selected', ascending=False)
    
    # return result, confusion_matrix, pathway_res, result1, result_pattern, result_pattern_mat
    return result, pathway_res, result1, result_pattern


# def extract_lr(adata, species, mean='algebra', min_cell=0, datahost='builtin'):
#     """
#     find overlapping LRs from CellChatDB
#     :param adata: AnnData object
#     :param species: support 'human', 'mouse' and 'zebrafish'
#     :param mean: 'algebra' (default) or 'geometric'
#     :param min_cell: for each selected pair, the spots expressing ligand or receptor should be larger than the min,
#     respectively.
#     :param datahost: the host of the ligand-receptor data. 'builtin' for package built-in otherwise from figshare
#     :return: ligand, receptor, geneInter (containing comprehensive info from CellChatDB) dataframes \
#             in adata.uns
#     """
#     if mean=='geometric':
#         from scipy.stats.mstats import gmean
#     adata.uns['mean'] = mean

#     if datahost == 'package':
#         if species in ['mouse', 'human', 'zerafish']:
#             datapath = './datasets/LR_data/%s-' %(species)
#         else:
#             raise ValueError("species type: {} is not supported currently. Please have a check.".format(species))
        
#         import pkg_resources
#         stream1 = pkg_resources.resource_stream(__name__, datapath + 'interaction_input_CellChatDB.csv.gz')
#         geneInter = pd.read_csv(stream1, index_col=0, compression='gzip')

#         stream2 = pkg_resources.resource_stream(__name__, datapath + 'complex_input_CellChatDB.csv')
#         comp = pd.read_csv(stream2, header=0, index_col=0)
#     else:
#         if species == 'mouse':
#             geneInter = pd.read_csv('https://figshare.com/ndownloader/files/36638919', index_col=0)
#             comp = pd.read_csv('https://figshare.com/ndownloader/files/36638916', header=0, index_col=0)
#         elif species == 'human':
#             geneInter = pd.read_csv('https://figshare.com/ndownloader/files/36638943', header=0, index_col=0)
#             comp = pd.read_csv('https://figshare.com/ndownloader/files/36638940', header=0, index_col=0)
#         elif species == 'zebrafish':
#             geneInter = pd.read_csv('https://figshare.com/ndownloader/files/38756022', header=0, index_col=0)
#             comp = pd.read_csv('https://figshare.com/ndownloader/files/38756019', header=0, index_col=0)
#         else:
#             raise ValueError("species type: {} is not supported currently. Please have a check.".format(species))
        
#     geneInter = geneInter.sort_values('annotation')
#     ligand = geneInter.ligand.values
#     receptor = geneInter.receptor.values
#     geneInter.pop('ligand')
#     geneInter.pop('receptor')

#     ## NOTE: the following for loop needs speed up
#     t = []
#     for i in range(len(ligand)):
#         for n in [ligand, receptor]:
#             l = n[i]
#             if l in comp.index:
#                 n[i] = comp.loc[l].dropna().values[pd.Series \
#                     (comp.loc[l].dropna().values).isin(adata.var_names)]
#             else:
#                 n[i] = pd.Series(l).values[pd.Series(l).isin(adata.var_names)]
#         if (len(ligand[i]) > 0) * (len(receptor[i]) > 0):
#             if mean=='geometric':
#                 meanL = gmean(adata[:, ligand[i]].X, axis=1)
#                 meanR = gmean(adata[:, receptor[i]].X, axis=1)
#             else:
#                 meanL = adata[:, ligand[i]].X.mean(axis=1)
#                 meanR = adata[:, receptor[i]].X.mean(axis=1)
#             if (sum(meanL > 0) >= min_cell) * \
#                     (sum(meanR > 0) >= min_cell):
#                 t.append(True)
#             else:
#                 t.append(False)
#         else:
#             t.append(False)
#     ind = geneInter[t].index
#     adata.uns['ligand'] = pd.DataFrame.from_records(zip_longest(*pd.Series(ligand[t]).values)).transpose()
#     adata.uns['ligand'].columns = ['Ligand' + str(i) for i in range(adata.uns['ligand'].shape[1])]
#     adata.uns['ligand'].index = ind
#     adata.uns['receptor'] = pd.DataFrame.from_records(zip_longest(*pd.Series(receptor[t]).values)).transpose()
#     adata.uns['receptor'].columns = ['Receptor' + str(i) for i in range(adata.uns['receptor'].shape[1])]
#     adata.uns['receptor'].index = ind
#     adata.uns['num_pairs'] = len(ind)
#     adata.uns['geneInter'] = geneInter.loc[ind]
#     if adata.uns['num_pairs'] == 0:
#         raise ValueError("No effective RL. Please have a check on input count matrix/species.")
#     return

# def spatialdm_global(adata, n_perm=1000, specified_ind=None, method='z-score', nproc=1):
#     """
#         global selection. 2 alternative methods can be specified.
#     :param n_perm: number of times for shuffling receptor expression for a given pair, default to 1000.
#     :param specified_ind: array containing queried indices for quick test/only run selected pair(s).
#     If not specified, selection will be done for all extracted pairs
#     :param method: default to 'z-score' for computation efficiency.
#         Alternatively, can specify 'permutation' or 'both'.
#         Two approaches should generate consistent results in general.
#     :param nproc: default to 1. Please decide based on your system.
#     :return: 'global_res' dataframe in adata.uns containing pair info and Moran p-values
#     """
#     if specified_ind is None:
#         specified_ind = adata.uns['geneInter'].index.values  # default to all pairs
#     else:
#         adata.uns['geneInter'] = adata.uns['geneInter'].loc[specified_ind]
#     total_len = len(specified_ind)
#     adata.uns['ligand'] = adata.uns['ligand'].loc[specified_ind]#.values
#     adata.uns['receptor'] = adata.uns['receptor'].loc[specified_ind]#.values
#     adata.uns['global_I'] = np.zeros(total_len)
#     adata.uns['global_stat'] = {}
#     if method in ['z-score', 'both']:
#         adata.uns['global_stat']['z']={}
#         adata.uns['global_stat']['z']['st'] = globle_st_compute(adata)
#         adata.uns['global_stat']['z']['z'] = np.zeros(total_len)
#         adata.uns['global_stat']['z']['z_p'] = np.zeros(total_len)
#     if method in ['both', 'permutation']:
#         adata.uns['global_stat']['perm']={}
#         adata.uns['global_stat']['perm']['global_perm'] = np.zeros((total_len, n_perm)).astype(np.float16)

#     if not (method in ['both', 'z-score', 'permutation']):
#         raise ValueError("Only one of ['z-score', 'both', 'permutation'] is supported")

#     with threadpool_limits(limits=nproc, user_api='blas'):
#         pair_selection_matrix(adata, n_perm, specified_ind, method)

#     adata.uns['global_res'] = pd.concat((adata.uns['ligand'], adata.uns['receptor']),axis=1)
#     # adata.uns['global_res'].columns = ['Ligand1', 'Ligand2', 'Ligand3', 'Receptor1', 'Receptor2', 'Receptor3', 'Receptor4']
#     if method in ['z-score', 'both']:
#         adata.uns['global_stat']['z']['z_p'] = np.where(np.isnan(adata.uns['global_stat']['z']['z_p']),
#                                                       1, adata.uns['global_stat']['z']['z_p'])
#         adata.uns['global_res']['z_pval'] = adata.uns['global_stat']['z']['z_p']
#         adata.uns['global_res']['z'] = adata.uns['global_stat']['z']['z']

#     if method in ['both', 'permutation']:
#         adata.uns['global_stat']['perm']['global_p'] = 1 - (adata.uns['global_I'] \
#                              > adata.uns['global_stat']['perm']['global_perm'].T).sum(axis=0) / n_perm
#         adata.uns['global_res']['perm_pval'] = adata.uns['global_stat']['perm']['global_p']
#     return

# def sig_pairs(adata, method='z-score', fdr=True, threshold=0.1):
#     """
#         select significant pairs
#     :param method: only one of 'z-score' or 'permutation' to select significant pairs.
#     :param fdr: True or False. If fdr correction will be done for p-values.
#     :param threshold: 0-1. p-value or fdr cutoff to retain significant pairs. Default to 0.1.
#     :return: 'selected' column in global_res containing whether or not a pair should be retained
#     """
#     adata.uns['global_stat']['method'] = method
#     if method == 'z-score':
#         _p = adata.uns['global_res']['z_pval'].values
#     elif method == 'permutation':
#         _p = adata.uns['global_res']['perm_pval'].values
#     else:
#         raise ValueError("Only one of ['z-score', 'permutation'] is supported")
#     if fdr:
#         _p = fdrcorrection(_p)[1]
#         adata.uns['global_res']['fdr'] = _p
#     adata.uns['global_res']['selected'] = (_p < threshold)

# def spatialdm_local(adata, n_perm=1000, method='z-score', specified_ind=None,
#                     nproc=1, scale_X=True):
#     """
#         local spot selection
#     :param n_perm: number of times for shuffling neighbors partner for a given spot, default to 1000.
#     :param method: default to 'z-score' for computation efficiency.
#         Alternatively, can specify 'permutation' or 'both' (recommended for spot number < 1000, multiprocesing).
#     :param specified_ind: array containing queried indices in sample pair(s).
#     If not specified, local selection will be done for all sig pairs
#     :param nproc: default to 1.
#     :return: 'local_stat' & 'local_z_p' and/or 'local_perm_p' in adata.uns.
#     """
#     adata.uns['local_stat'] = {}
#     if (int(n_perm / nproc) != (n_perm / nproc)):
#         raise ValueError("n_perm should be divisible by nproc")
#     if type(specified_ind) == type(None):
#         specified_ind = adata.uns['global_res'][
#             adata.uns['global_res']['selected']].index  # default to global selected pairs
#     # total_len = len(specified_ind)
#     ligand = adata.uns['ligand'].loc[specified_ind]
#     receptor = adata.uns['receptor'].loc[specified_ind]
#     ind = ligand.index
#     adata.uns['local_stat']['local_I'] = np.zeros((adata.shape[0], len(ind)))
#     adata.uns['local_stat']['local_I_R'] = np.zeros((adata.shape[0], len(ind)))
#     N = adata.shape[0]
#     if method in ['both', 'permutation']:
#         adata.uns['local_stat']['local_permI'] = np.zeros((len(ind), n_perm, N))
#         adata.uns['local_stat']['local_permI_R'] = np.zeros((len(ind), n_perm, N))
#     if method in ['both', 'z-score']:
#         adata.uns['local_z'] = np.zeros((len(ind), adata.shape[0]))
#         adata.uns['local_z_p'] = np.zeros((len(ind), adata.shape[0]))

#     ## different approaches
#     with threadpool_limits(limits=nproc, user_api='blas'):
#         spot_selection_matrix(adata, ligand, receptor, ind, n_perm, method, scale_X)


# def sig_spots(adata, method='z-score', fdr=True, threshold=0.1):
#     """
#         pick significantly co-expressing spots
#     :param method: one of the methods from spatialdm_local, default to 'z-score'.
#     :param fdr: True or False, default to True
#     :param threshold: p-value or fdr cutoff to retain significant pairs. Default to 0.1.
#     :return:  1) 'selected_spots' in adata.uns: a binary frame of which spots being selected for each pair;
#      2) 'n_spots' in adata.uns['local_stat']: number of selected spots for each pair.
#     """
#     if method == 'z-score':
#         _p = adata.uns['local_z_p']
#     if method == 'permutation':
#         _p = adata.uns['local_perm_p']
#     if fdr:
#         _fdr = fdrcorrection(np.hstack(_p.values))[1].reshape(_p.shape)
#         _p.loc[:,:] = _fdr
#         adata.uns['local_stat']['local_fdr'] = _p
#     adata.uns['selected_spots'] = (_p < threshold)
#     adata.uns['local_stat']['n_spots'] = adata.uns['selected_spots'].sum(1)
#     adata.uns['local_stat']['local_method'] = method
#     return

# def drop_uns_na(adata, global_stat=False, local_stat=False):
#     adata.uns['geneInter'] = adata.uns['geneInter'].fillna('NA')
#     adata.uns['global_res'] = adata.uns['global_res'].fillna('NA')
#     adata.uns['ligand'] = adata.uns['ligand'].fillna('NA')
#     adata.uns['receptor'] = adata.uns['receptor'].fillna('NA')
#     adata.uns['local_stat']['n_spots'] = pd.DataFrame(adata.uns['local_stat']['n_spots'], columns=['n_spots'])
#     if global_stat and ('global_stat' in adata.uns.keys()):
#         adata.uns.pop('global_stat')
#     if local_stat and ('local_stat' in adata.uns.keys()):
#         adata.uns.pop('local_stat')

# def restore_uns_na(adata):
#     adata.uns['geneInter'] = adata.uns['geneInter'].replace('NA', np.nan)
#     adata.uns['global_res'] = adata.uns['global_res'].replace('NA', np.nan)
#     adata.uns['ligand'] = adata.uns['ligand'].replace('NA', np.nan)
#     adata.uns['receptor'] = adata.uns['receptor'].replace('NA', np.nan)
#     adata.uns['local_stat']['n_spots'] =  adata.uns['local_stat']['n_spots'].n_spots

# def write_spatialdm_h5ad(adata, filename=None):
#     if filename is None:
#         filename = 'spatialdm_out.h5ad'
#     elif not filename.endswith('h5ad'):
#         filename = filename+'.h5ad'
#     drop_uns_na(adata)
#     adata.write(filename)

# def read_spatialdm_h5ad(filename):
#     adata = ann.read_h5ad(filename)
#     restore_uns_na(adata)
#     return adata





# add shortcuts to the package's first level

from . import datasets
# from . import HIPT
# from .main import *
from .version import __version__

from .loadData import *
from .processData import *
from .model import *
from .traintest import *
from .inference import *
from .evaluation import * 
from .SparseAEH import *
from .plottings import *
from .cropping import *
from .downstream import *




from copy import copy
from ._util_dataset import AMetadata


_NPC = AMetadata(
    name="NPC",
    doc_header="Nasopharyngeal carcinoma (NPC) dataset from `Gong et al <https://doi.org/10.1038/s41467-023-37614-6>`__.",
    shape=(1331, 36601),
    url="https://figshare.com/ndownloader/files/48619396", 
)

_BRCA_FineST_allspot = AMetadata(
    name="BRCA_FineST_allspot",
    doc_header="Breast cancer (BRCA) dataset (all spots) from `Janesick et al <https://doi/10.1038/s41467-023-43458-x`__.",
    shape=(19732, 863),
    url="https://figshare.com/ndownloader/files/49070155",
)

_BRCA_Original_whole = AMetadata(
    name="BRCA_Original_whole",
    doc_header="Breast cancer (BRCA) dataset (all spots) from `Janesick et al <https://doi/10.1038/s41467-023-43458-x`__.",
    shape=(4992, 18085),
    url="https://figshare.com/ndownloader/files/49286560",
)

_BRCA_FineST_subspot = AMetadata(
    name="BRCA_FineST_subspot",
    doc_header="Breast cancer (BRCA) dataset from `Janesick et al <https://doi/10.1038/s41467-023-43458-x`__.",
    shape=(4992, 18085),
    url="https://ndownloader.figshare.com/files/40178041",
)

_CRC16um = AMetadata(
    name="CRC16um",
    doc_header="Original colorectal cancer (CRC) dataset from Oliveira, et al. <https://www.biorxiv.org/content/10.1101/2024.06.04.597233v1.full>`__.",
    shape=(137051, 18085),  
    url="https://figshare.com/ndownloader/files/49633644", 
)

_CRC08um = AMetadata(
    name="CRC08um",
    doc_header="Original colorectal cancer (CRC) dataset from Oliveira, et al. <https://www.biorxiv.org/content/10.1101/2024.06.04.597233v1.full>`__.",
    shape=(545913, 18085),  
    url="https://figshare.com/ndownloader/files/50571282", 
)

_CRC_Original = AMetadata(
    name="CRC_Original",
    doc_header="Original colorectal cancer (CRC) dataset from Oliveira, et al. <https://www.biorxiv.org/content/10.1101/2024.06.04.597233v1.full>`__.",
    shape=(136954, 862), #(137051, 18085),  
    url="https://figshare.com/ndownloader/files/48724144", 
)

_CRC_FineST = AMetadata(
    name="CRC_FineST",
    doc_header="FineST colorectal cancer (CRC) dataset  from Oliveira, et al. <https://www.biorxiv.org/content/10.1101/2024.06.04.597233v1.full>`__.",
    shape=(136954, 862),
    url="https://figshare.com/ndownloader/files/48724447",
)

_A3 = AMetadata(
    name="A3",
    doc_header="12-PCW Fetus colon single rep from Corbett, et al. <https://doi.org/10.1016/j.cell.2020.12.016>`__.",
    shape=(1080, 33538),
    url="https://figshare.com/ndownloader/files/40178311",
)

_A4 = AMetadata(
    name="A4",
    doc_header="19-PCW Fetus colon single rep from Corbett, et al. <https://doi.org/10.1016/j.cell.2020.12.016>`__.",
    shape=(1242, 33538),
    url="https://figshare.com/ndownloader/files/40178314",
)

_A6 = AMetadata(
    name="A6",
    doc_header="12-PCW TI rep 1 from Corbett, et al. <https://doi.org/10.1016/j.cell.2020.12.016>`__.",
    shape=(346, 33538),
    url="https://figshare.com/ndownloader/files/40178017",
)

_A7 = AMetadata(
    name="A7",
    doc_header="12-PCW TI rep 2 from Corbett, et al. <https://doi.org/10.1016/j.cell.2020.12.016>`__.",
    shape=(344, 33538),
    url="https://figshare.com/ndownloader/files/40178014",
)

_A8 = AMetadata(
    name="A8",
    doc_header="12-PCW colon rep 1 from Corbett, et al. <https://doi.org/10.1016/j.cell.2020.12.016>`__.",
    shape=(709, 33538),
    url="https://figshare.com/ndownloader/files/40178011",
)

_A9 = AMetadata(
    name="A9",
    doc_header="12-PCW colon rep 2 from Corbett, et al. <https://doi.org/10.1016/j.cell.2020.12.016>`__.",
    shape=(644, 33538),
    url="https://figshare.com/ndownloader/files/40178308",
)


for name, var in copy(locals()).items():
    if isinstance(var, AMetadata):
        var._create_function(name, globals())


__all__ = [  # noqa: F822
    "NPC", "BRCA_FineST_allspot", "BRCA_Original_whole", "BRCA_FineST_subspot", "CRC16um", "CRC08um", "CRC_Original", "CRC_FineST", "A3","A4","A6","A7","A8","A9"]



import torch
import numpy as np
from .utils import *
from .loadData import * 


def adjust_learning_rate(optimizer, epoch, initial_lr, num_epochs):
    
    """Adjusts the learning rate based on the cosine annealing strategy."""
    
    lr = 0.5 * initial_lr * (1 + np.cos(np.pi * epoch / num_epochs))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr
    return lr


def save_model(model, dir_name, params, optimizer, LOSS):
    cur_save_path = os.path.join(dir_name, "epoch_"+str(params["training_epoch"])+".pt")
    torch.save({
                  'epoch': params['training_epoch'],
                  'model_state_dict': model.state_dict(),
                  'optimizer_state_dict': optimizer.state_dict(),
                  'loss': LOSS,
                  'params': params,
                  # 'train_genes':train_genes,
                   }, cur_save_path)
    

    
# def train_model(params, model, train_loader, optimizer, cur_epoch, l, tree_type, leaf_size, dataset_class): 
    
#     print("train model")    
    
#     cur_lr = adjust_learning_rate(optimizer, cur_epoch, params['inital_learning_rate'], params['training_epoch'])
#     total_loss, total_num = 0.0, 0.0

#     ## load data
#     (cur_train_data_mat, 
#      cur_train_matrix_mat, 
#      cur_train_coors_mat, 
#      cur_pos_info) = loadTrainTestData(train_loader, neighbor_k=params['k_nearest_positives'], 
#                                        tree_type=tree_type, leaf_size=leaf_size, dataset_class=dataset_class)

    
#     for image_profile, gene_profile, positive_index, _ in loadBatchData(cur_train_data_mat, 
#                                                                         cur_train_matrix_mat,
#                                                                         cur_train_coors_mat, 
#                                                                         params['batch_size_pair'],
#                                                                         cur_pos_info):
        
#         input_gene_exp = torch.tensor(np.asarray(gene_profile)).float().to(device)   # torch.Size([64, 128])
#         image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [1331, 256, 384] --> [1331*256, 384]
#         input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
        
#         ## model
#         (representation_matrix, 
#          reconstruction_matrix, 
#          projection_matrix,
#          representation_image, 
#          reconstruction_iamge, 
#          projection_image) = model(input_gene_exp, input_image_exp)     
        
#         ## reshape
#         _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
#         _, projection_image_reshape = reshape_latent_image(projection_image, dataset_class)
    
#         ## cross decoder
#         reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
#         _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)

#         ## compute the loss
#         loss = l(
#             projection_image_reshape,
#             projection_matrix,
#             # representation_image_reshape,
#             # representation_matrix,
#             torch.tensor(positive_index).to(device),
#             input_image_exp,
#             reconstruction_iamge,
#             reconstruction_matrix,
#             reconstruction_iamge_reshapef2,
#             # input_gene_exp,
#             # w1=1, w2=1, w3=1, w4=1
#             input_gene_exp
#         )  

#         # optimization
#         optimizer.zero_grad()
#         loss.backward()
#         optimizer.step()

#         total_loss += loss.item() * gene_profile.shape[0]
#         total_num += gene_profile.shape[0]
#         pass
    
#     LOSS = total_loss/total_num    # shumin end

#     return LOSS



# def test_model(params, model, test_loader, l, tree_type, leaf_size, dataset_class): 
    
#     print("test model")
    
#     # load data
#     (cur_train_data_mat, 
#      cur_train_matrix_mat, 
#      cur_train_coors_mat, 
#      cur_pos_info) = loadTrainTestData(test_loader, neighbor_k=params['k_nearest_positives'], 
#                                        tree_type=tree_type, leaf_size=leaf_size, dataset_class=dataset_class)
        
#     total_loss, total_num = 0.0, 0.0
#     for image_profile, gene_profile, positive_index, _ in loadBatchData(cur_train_data_mat, 
#                                                                         cur_train_matrix_mat,
#                                                                         cur_train_coors_mat, 
#                                                                         params['batch_size_pair'],
#                                                                         cur_pos_info):
        
#         input_gene_exp = torch.tensor(np.asarray(gene_profile)).float().to(device)   # torch.Size([64, 128])
#         image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [1331, 256, 384] --> [1331*256, 384]
#         input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
        
#         ## model
#         (representation_matrix, 
#          reconstruction_matrix, 
#          projection_matrix,
#          representation_image, 
#          reconstruction_iamge, 
#          projection_image) = model(input_gene_exp, input_image_exp)     
        
#         ## reshape
#         _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
#         _, projection_image_reshape = reshape_latent_image(projection_image, dataset_class)
    
#         ## cross decoder
#         reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
#         _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)

#         ## compute the loss
#         loss = l(
#             projection_image_reshape,
#             projection_matrix,
#             # representation_image_reshape,
#             # representation_matrix,
#             torch.tensor(positive_index).to(device),
#             input_image_exp,
#             reconstruction_iamge,
#             reconstruction_matrix,
#             reconstruction_iamge_reshapef2,
#             # input_gene_exp,
#             # w1=1, w2=1, w3=1, w4=1
#             input_gene_exp
#         )  
    
#         total_loss += loss.item() * gene_profile.shape[0]
#         total_num += gene_profile.shape[0]
#         pass
    
#     LOSS = total_loss/total_num 
    
#     # count = batch["reduced_expression"].size(0)
#     # loss_meter.update(loss.item(), count)
#     # tqdm_object.set_postfix(valid_loss=loss_meter.avg)

#     return LOSS



##################################
# 2025.01.06 Adjust loss function
# Train and Test model adjust
##################################
def train_model(params, model, train_loader, optimizer, cur_epoch, l, tree_type, leaf_size, dataset_class): 
    
    print("train model")    
    
    cur_lr = adjust_learning_rate(optimizer, cur_epoch, params['inital_learning_rate'], params['training_epoch'])
    total_loss, total_num = 0.0, 0.0

    ## load data
    (cur_train_data_mat, 
     cur_train_matrix_mat, 
     cur_train_coors_mat, 
     cur_pos_info) = loadTrainTestData(train_loader, neighbor_k=params['k_nearest_positives'], 
                                       tree_type=tree_type, leaf_size=leaf_size, dataset_class=dataset_class)

    
    for image_profile, gene_profile, positive_index, _ in loadBatchData(cur_train_data_mat, 
                                                                        cur_train_matrix_mat,
                                                                        cur_train_coors_mat, 
                                                                        params['batch_size_pair'],
                                                                        cur_pos_info):
        
        input_gene_exp = torch.tensor(np.asarray(gene_profile)).float().to(device)   # torch.Size([64, 128])
        image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [1331, 256, 384] --> [1331*256, 384]
        input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
        
        ## model
        (representation_matrix, 
         reconstruction_matrix, 
         projection_matrix,
         representation_image, 
         reconstruction_iamge, 
         projection_image,
         recon_mat_mean, recon_mat_disp, recon_mat_pi) = model(input_gene_exp, input_image_exp)     
        
        ## reshape
        _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
        _, projection_image_reshape = reshape_latent_image(projection_image, dataset_class)
    
        ## cross decoder
        reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
        _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)

        ## compute the loss
        loss = l(
            projection_image_reshape,
            projection_matrix,
            # representation_image_reshape,
            # representation_matrix,
            torch.tensor(positive_index).to(device),
            input_image_exp,
            reconstruction_iamge,
            reconstruction_matrix,
            reconstruction_iamge_reshapef2,
            # input_gene_exp,
            # w1=1, w2=1, w3=1, w4=1
            input_gene_exp,
            recon_mat_mean, recon_mat_disp, recon_mat_pi
        )  

        # optimization
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        total_loss += loss.item() * gene_profile.shape[0]
        total_num += gene_profile.shape[0]
        pass
    
    LOSS = total_loss/total_num    # shumin end

    return LOSS



def test_model(params, model, test_loader, l, tree_type, leaf_size, dataset_class): 
    
    print("test model")
    
    # load data
    (cur_train_data_mat, 
     cur_train_matrix_mat, 
     cur_train_coors_mat, 
     cur_pos_info) = loadTrainTestData(test_loader, neighbor_k=params['k_nearest_positives'], 
                                       tree_type=tree_type, leaf_size=leaf_size, dataset_class=dataset_class)
        
    total_loss, total_num = 0.0, 0.0
    for image_profile, gene_profile, positive_index, _ in loadBatchData(cur_train_data_mat, 
                                                                        cur_train_matrix_mat,
                                                                        cur_train_coors_mat, 
                                                                        params['batch_size_pair'],
                                                                        cur_pos_info):
        
        input_gene_exp = torch.tensor(np.asarray(gene_profile)).float().to(device)   # torch.Size([64, 128])
        image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [1331, 256, 384] --> [1331*256, 384]
        input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
        
        ## model
        (representation_matrix, 
         reconstruction_matrix, 
         projection_matrix,
         representation_image, 
         reconstruction_iamge, 
         projection_image,
         recon_mat_mean, recon_mat_disp, recon_mat_pi) = model(input_gene_exp, input_image_exp)     
        
        ## reshape
        _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
        _, projection_image_reshape = reshape_latent_image(projection_image, dataset_class)
    
        ## cross decoder
        reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
        _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)

        ## compute the loss
        loss = l(
            projection_image_reshape,
            projection_matrix,
            # representation_image_reshape,
            # representation_matrix,
            torch.tensor(positive_index).to(device),
            input_image_exp,
            reconstruction_iamge,
            reconstruction_matrix,
            reconstruction_iamge_reshapef2,
            # input_gene_exp,
            # w1=1, w2=1, w3=1, w4=1
            input_gene_exp,
            recon_mat_mean, recon_mat_disp, recon_mat_pi
        )  
    
        total_loss += loss.item() * gene_profile.shape[0]
        total_num += gene_profile.shape[0]
        pass
    
    LOSS = total_loss/total_num 
    
    # count = batch["reduced_expression"].size(0)
    # loss_meter.update(loss.item(), count)
    # tqdm_object.set_postfix(valid_loss=loss_meter.avg)

    return LOSS


## 2024.10.31 Create by Tianjie Wang (our co-author)
##            https://github.com/jackywangtj66/SparseAEH
## 2024.10.31 intergrate base.py and plot.py together

import numpy as np
#from function import RBF_kernel
import scipy
import time
from sklearn.metrics.pairwise import laplacian_kernel,rbf_kernel
from scipy.sparse import csr_matrix
from operator import itemgetter
import warnings
from sklearn.cluster import KMeans
from scipy.spatial import distance_matrix
import random

def _pinv_1d(v, eps=1e-5):
    return np.array([0 if abs(x) <= eps else 1/x for x in v], dtype=float)

def _pexp(array):
    return np.array([np.inf if n>50 else (np.exp(n) if n >-50 else 0) for n in array])


class Kernel:

    def __init__(self,spatial,ss_loc=None,group_size=16,cov=None,dependency=None,
                 d=5,kernel='laplacian',l=0.01):
        """
        number of superspots: M
        cov: full pre-determined covariance matrix
        dependency format: M-length list of lists, 
        the i-th element indicates the superspots that the i-th superspot is dependent on
        ss_loc: M-length list of lists, the i-th element indicates the spots (ordinal) that 
        the i-th superspots contain spatial: spatial coordinates of all spots
        l: hyperparameter for kernel
        """
        super().__init__()
        self.N = len(spatial)
        #self.all_cov = [[0 for _ in range(self.M)] for _ in range(self.M)]
        #self.ss_loc = ss_loc
        self.spatial = spatial
        self.cond_mean = None
        self.l = l
        self.d = d
        self.A = []
        self.cond_cov = []
        self.kernel = kernel
        self._initialize(cov,ss_loc,dependency,group_size)
    
    def get_mat(self,rows,cols):
        row = []
        for r in rows:
            col = []
            for c in cols:
                if r >= c:
                    col.append(self.all_cov[r][c])
                else:
                    col.append(self.all_cov[c][r].T)
            row.append(np.hstack(col))
        return np.vstack(row)

    
    def _initialize(self,cov,ss_loc,dependency,group_size):
        self._init_ss(ss_loc,dependency,group_size)
        self._init_ds_loc()
        self._init_all_cov(cov)
        self._init_ds_eig()
        self._init_base_cond_cov()        
        #self._init_cond_cov()

    def _init_ss(self,ss_loc,dependency,group_size):
        if ss_loc is not None:
            self.ss_loc = ss_loc
            self.dependency = dependency
            self.M = len(dependency)
        else:
            centers = []
            self.ss_loc = []
            group = self.N / group_size
            g_r = int(np.sqrt(group))
            g_c = int(group/g_r)
            #self.M = g_c * g_r
            kmeans_r = KMeans(n_clusters=g_r, random_state=0, n_init=10)
            kmeans_c = KMeans(n_clusters=g_c, random_state=0, n_init=10)
            kmeans_r.fit(self.spatial[:,0:1])
            kmeans_c.fit(self.spatial[:,1:])
            for r in range(g_r):
                for c in range(g_c):
                    pos = np.logical_and(kmeans_r.labels_ == r, kmeans_c.labels_ == c)
                    if pos.any():
                        centers.append(np.average(self.spatial[pos],axis=0))
                        self.ss_loc.append(np.arange(len(self.spatial))[pos])
            #print(self.ss_loc)
            #print(centers)
            centers = np.vstack(centers)
            self.M = centers.shape[0]
            self.dependency = [[] for _ in range(self.M)]
            distance = distance_matrix(centers,centers)
            for i in range(self.M):
                if i <= self.d:
                    self.dependency[i] = list(range(i))
                else:
                    #self.dependency[i] = np.argmax(distance[i,:i],self.d)
                    self.dependency[i] = distance[i,:i].argsort()[:self.d]


    def _init_ds_loc(self):
        self.ds_loc = []        
        for i,ds in enumerate(self.dependency):
           ind = []
           for ss in ds:
               ind += np.array(self.ss_loc[ss]).tolist()
           self.ds_loc.append(ind)
    
    def _init_all_cov(self,cov):
        self.all_cov = [[0 for _ in range(self.M)] for _ in range(self.M)]
        for i in range(self.M):
            if cov is not None:
                self.all_cov[i][i] = cov[np.ix_(self.ss_loc[i],self.ss_loc[i])]
            else:
                if self.kernel == 'rbf':
                    self.all_cov[i][i] = rbf_kernel(self.spatial[self.ss_loc[i]],gamma=self.l)
                else:
                    self.all_cov[i][i] = laplacian_kernel(self.spatial[self.ss_loc[i]],gamma=self.l)
            for j in self.dependency[i]:
                if cov is not None:
                    self.all_cov[i][j] = cov[np.ix_(self.ss_loc[i],self.ss_loc[j])]
                else:
                    if self.kernel == 'rbf':
                        self.all_cov[i][j] = rbf_kernel(self.spatial[self.ss_loc[i]],
                                                        self.spatial[self.ss_loc[j]],gamma=self.l)
                    else:
                        self.all_cov[i][j] = laplacian_kernel(self.spatial[self.ss_loc[i]],
                                                              self.spatial[self.ss_loc[j]],gamma=self.l)
        for k in range(self.M):
            for i in self.dependency[k]:
                for j in self.dependency[k]:
                    if j<i and isinstance(self.all_cov[i][j],int):
                        if cov is not None:
                            self.all_cov[i][j] = cov[np.ix_(self.ss_loc[i],self.ss_loc[j])]
                        else:
                            if self.kernel == 'rbf':
                                self.all_cov[i][j] = rbf_kernel(self.spatial[self.ss_loc[i]],
                                                                self.spatial[self.ss_loc[j]],gamma=self.l)
                            else:
                                self.all_cov[i][j] = laplacian_kernel(self.spatial[self.ss_loc[i]],
                                                                      self.spatial[self.ss_loc[j]],
                                                                      gamma=self.l)
    

    def _init_ds_eig(self):
        #C_m,C_m
        self.ds_eig = []
        for i in range(self.M):
            if len(self.dependency[i]) == 0:
                self.ds_eig.append(())
                self.A.append(())            
            else:            
                ds_cov = self.get_mat(self.dependency[i],self.dependency[i])
                s,u = np.linalg.eigh(ds_cov)
                #s_inv = _pinv_1d(s)
                self.ds_eig.append((s,u))
                self.A.append(self.get_mat([i],self.dependency[i]) @ u)
    

    def _init_base_cond_cov(self):
        for i in range(self.M):
            if len(self.dependency[i]) == 0:
                self.cond_cov.append(self.get_mat([i],[i]))
            else:
                self.cond_cov.append(self.get_mat([i],[i]) - np.multiply(1/self.ds_eig[i][0],self.A[i])@self.A[i].T)
    

    def update_cond_cov(self,Delta):
        cond_cov_eig = [[] for _ in range(len(Delta))]
        for eig, delta in zip(cond_cov_eig,Delta):
            for i in range(self.M):
                if len(self.dependency[i]) == 0:
                #self.cond_cov.append(self.kernel.base_cond_cov[i]+self.delta*np.eye(len(self.kernel.ss_loc[i])))
                    s,u = np.linalg.eigh(self.cond_cov[i]+delta*np.eye(len(self.ss_loc[i])))
                else:
                    s,u = np.linalg.eigh(self.cond_cov[i]+delta*np.eye(len(self.ss_loc[i]))+
                                         delta*np.multiply(1/((self.ds_eig[i][0]+delta)*self.ds_eig[i][0]),self.A[i])@self.A[i].T)
                eig.append((s,u))
        return cond_cov_eig

class MixedGaussian:
    def __init__(self,spatial,ss_loc=None,group_size=16,
                 cov=None,dependency=None,d=5,kernel='rbf',l=0.01):
        self.kernel = Kernel(spatial,ss_loc,group_size,cov,dependency,d,kernel,l)

    def update_cond_mean(self):
        #Y:N*G  mean:N*K
        self.cond_dev = self.Y[np.newaxis,:] - self.mean.transpose()[:,:,np.newaxis]   #K,N,G
        dev = self.cond_dev.copy()
        for i in range(self.kernel.M):
                if len(self.kernel.dependency[i]) > 0:
                    for k in range(self.K):
                        self.cond_dev[k,self.kernel.ss_loc[i],:] = self.cond_dev[k,self.kernel.ss_loc[i],:] - \
                        np.multiply(1/(self.kernel.ds_eig[i][0]+self.delta[k]),
                                    self.kernel.A[i]) @ self.kernel.ds_eig[i][1].T @ dev[k,self.kernel.ds_loc[i],:]

    def compute_ll(self,cond_cov_eig):
        ll = np.zeros((self.G,self.K))
        for k in range(self.K):
            ll[:,k] = np.log(2 * np.pi)*self.N + 2*np.log(self.sigma_sq[k])*self.N
            for i in range(self.kernel.M):
                det = np.prod(cond_cov_eig[k][i][0])
                if det <= 0:
                    print(cond_cov_eig[k][i][0]) 
                ll[:,k] += np.log(det)
                temp = self.cond_dev[k][self.kernel.ss_loc[i],:].T @ cond_cov_eig[k][i][1]
                ll[:,k] += np.sum(np.multiply(1/cond_cov_eig[k][i][0],np.square(temp)),
                                  axis=1)/self.sigma_sq[k]
        ll = ll*-0.5
        return ll
    
    def update_param(self,omega):
        new_mean= np.zeros_like(self.mean)
        #mean
        for k in range(self.K):
            new_mean[:,k:(k+1)] = self.Y @ omega[:,k:(k+1)] / np.sum(omega[:,k])
        new_dev = self.Y[np.newaxis,:] - new_mean.transpose()[:,:,np.newaxis]
        #pi
        if self.update_pi:
            self.pi = np.average(omega,axis=0)
        
        for k in range(self.K):
            self.cov_new = []
            for i in range(self.kernel.M):
                l = len(self.kernel.ss_loc[i])
                cov_i = np.zeros((l,l))
                for g in range(self.G):
                    cov_i += omega[g,k]*np.outer(new_dev[k,self.kernel.ss_loc[i],g],
                                                 new_dev[k,self.kernel.ss_loc[i],g])
                cov_i = cov_i/np.sum(omega[:,k])
                self.cov_new.append(cov_i)
            numer,t_2 = 0,0
            for i in range(self.kernel.M):
                numer += np.sum(np.multiply(self.kernel.all_cov[i][i],self.cov_new[i]))
                #denom += np.sum(np.multiply(self.kernel.all_cov[i][i],self.kernel.all_cov[i][i]))
                #t_1 += np.trace(self.kernel.all_cov[i][i])
                t_2 += np.trace(self.cov_new[i])
            #print(numer,denom,t_1,t_2)
            self.sigma_sq[k] = (numer - self.t_1*t_2/self.N)*0.5 / (self.denom - self.t_1**2/self.N) + self.sigma_sq[k]/2
            #print(self.sigma_sq[k],(numer - t_1*t_2/self.N) / (denom - t_1**2/self.N))
            if self.sigma_sq[k] == 0:
                self.delta[k] = t_2*0.5/self.N + self.delta[k]/2
            else:
                self.delta[k] = (t_2-self.sigma_sq[k]*self.t_1)*0.5/(self.N*self.sigma_sq[k]) + self.delta[k]/2
            if self.delta[k]<=0:
                self.delta[k] = 0 
                self.sigma_sq[k] = numer / self.denom 
        return new_mean
    
    def update_mean(self,omega):
        new_mean= np.zeros_like(self.mean)
        #mean
        for k in range(self.K):
            new_mean[:,k:(k+1)] = self.Y @ omega[:,k:(k+1)] / np.sum(omega[:,k])
        new_dev = self.Y[np.newaxis,:] - new_mean.transpose()[:,:,np.newaxis]
        #pi
        if self.update_pi:
            self.pi = np.average(omega,axis=0)
        return new_mean
    
    def param_init(self):
        samp_ind = random.sample(range(self.G),self.G//10)
        sample = self.Y[:,samp_ind]
        kmeans = KMeans(n_clusters=self.K, random_state=0).fit(sample.T)
        return kmeans.cluster_centers_.T

    def run_cluster(self,Y,K,pi=None,mean=None,sigma_sq=None,delta=None,iter=500,
                    threshold=5e-2,init_mean='k_means',update_pi=True):
        self.Y = Y
        self.K = K
        self.N,self.G = self.Y.shape
        self.update_pi = update_pi
        if pi is not None:
            self.pi = np.array(pi,dtype=float)
        else:
            #self.pi = np.random.dirichlet(np.ones(self.K))    
            self.pi = np.ones(self.K,dtype=float) /self.K
        if mean is not None:
            self.mean = np.array(mean,dtype=float)
        else:    
            #self.mean = np.abs(np.random.normal(size=(self.N, self.K)))
            self.mean = np.random.uniform(size=(self.N, self.K))
            if init_mean == 'k_means':
                self.init_mean = self.param_init()
                self.mean = self.init_mean
            elif init_mean == 'sample':
                self.init_mean = self.Y[:,np.random.choice(self.G,self.K)]
            elif isinstance(init_mean,np.ndarray):
                self.init_mean = init_mean
            
        #return self.mean
        if sigma_sq is not None:
            self.sigma_sq = np.array(sigma_sq,dtype=float)
        else:
            self.sigma_sq = np.ones(self.K,dtype=float)*0.1
        if delta is not None:
            self.delta = np.array(delta,dtype=float)
        else:
            self.delta = np.ones(self.K,dtype=float)*1
        
        # power = np.zeros((G,self.K))
        self.omega = np.ones((self.G,self.K))/self.K
        converge = False
        count = 0
        #self.ll = np.zeros((self.G,self.K))
        self.denom,self.t_1 = 0,0
        for i in range(self.kernel.M):
            self.denom += np.sum(np.multiply(self.kernel.all_cov[i][i],self.kernel.all_cov[i][i]))
            self.t_1 += np.trace(self.kernel.all_cov[i][i])
        
        while not converge:
            print('Iteration {}'.format(count))
            cond_cov_eig = self.kernel.update_cond_cov(self.delta)
            self.update_cond_mean()

            self.ll = self.compute_ll(cond_cov_eig)
            #print(self.ll)
            #print(compute_likelihood(self.Y,self.kernel,self.cond_dev[0],self.sigma_sq[0],cond_cov_eig[0]))
            for k in range(self.K):
                if self.pi[k] == 0:
                    self.omega[:,k] = 0
                else:
                    for g in range(self.G):
                        #omega[g,k] = 3/4*self.pi[k]/np.sum(self.pi * _pexp((self.ll[g]-self.ll[g][k])/np.sqrt(self.N))) + omega[g,k]/4
                        self.omega[g,k] = self.pi[k]/np.sum(self.pi * _pexp((self.ll[g]-self.ll[g][k])/np.sqrt(self.N))) + 1e-3/self.G
                        # if np.sum(self.pi * _pexp((ll[g]-ll[g][k])/np.sqrt(self.N))) == 0:
                        #     print(ll[g],ll[g][k])

            #print(self.omega)
            new_mean = self.update_mean(self.omega)
            #print(self.delta,self.sigma_sq)
            count += 1
            if count > iter or np.mean(np.abs(new_mean-self.mean))<0.1:
                converge = True
            self.mean = new_mean
            #converge = True
            #print(self.pi,self.sigma_sq,self.delta)
            #indexes = np.array([np.arange(0,2),np.arange(300,302),np.arange(800,802)])
            #print(self.sigma_sq,self.delta,self.pi,omega[indexes])
        #return self.mean
        print('updating variance')
        converge = False
        while not converge:
            print('Iteration {}'.format(count))
            cond_cov_eig = self.kernel.update_cond_cov(self.delta)
            self.update_cond_mean()

            self.ll = self.compute_ll(cond_cov_eig)
            #print(self.ll)
            #print(compute_likelihood(self.Y,self.kernel,self.cond_dev[0],self.sigma_sq[0],cond_cov_eig[0]))
            for k in range(self.K):
                if self.pi[k] == 0:
                    self.omega[:,k] = 0
                else:
                    for g in range(self.G):
                        #omega[g,k] = 3/4*self.pi[k]/np.sum(self.pi * _pexp((self.ll[g]-self.ll[g][k])/np.sqrt(self.N))) + omega[g,k]/4
                        self.omega[g,k] = self.pi[k]/np.sum(self.pi * _pexp((self.ll[g]-self.ll[g][k]))) + 1e-3/self.G
                        # if np.sum(self.pi * _pexp((ll[g]-ll[g][k]))) == 0:
                        #     print(ll[g],ll[g][k])

            #print(self.omega)
            new_mean = self.update_param(self.omega)
            #print(self.delta,self.sigma_sq)
            count += 1
            if count > iter or np.mean(np.abs(new_mean-self.mean))<threshold:
                converge = True
            self.mean = new_mean
            #print(self.pi,self.sigma_sq,self.delta)
        self.labels = np.argmax(self.omega,axis=1)    
        return self.mean

    def cluster_counts(self,query_label=None):
        if query_label is not None:
            return np.sum(self.labels==query_label)
        else:
            return np.array([np.sum(self.labels==i) for i in range(self.K)])


import torch
import numpy as np
from .utils import *
from .loadData import * 


def adjust_learning_rate(optimizer, epoch, initial_lr, num_epochs):
    """Adjusts the learning rate based on the cosine annealing strategy."""
    lr = 0.5 * initial_lr * (1 + np.cos(np.pi * epoch / num_epochs))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr
    return lr


def save_model(model, dir_name, params, optimizer, LOSS):
    cur_save_path = os.path.join(dir_name, "epoch_"+str(params["training_epoch"])+".pt")
    torch.save({
                  'epoch': params['training_epoch'],
                  'model_state_dict': model.state_dict(),
                  'optimizer_state_dict': optimizer.state_dict(),
                  'loss': LOSS,
                  'params': params,
                  # 'train_genes':train_genes,
                   }, cur_save_path)
    
    
def train_model(params, model, train_loader, optimizer, cur_epoch, l, tree_type, leaf_size, dataset_class): 
    
    print("train model")    
    
    cur_lr = adjust_learning_rate(optimizer, cur_epoch, params['inital_learning_rate'], params['training_epoch'])
    total_loss, total_num = 0.0, 0.0

    ## load data
    (cur_train_data_mat, 
     cur_train_matrix_mat, 
     cur_train_coors_mat, 
     cur_pos_info) = loadTrainTestData(train_loader, neighbor_k=params['k_nearest_positives'], 
                                       tree_type=tree_type, leaf_size=leaf_size, dataset_class=dataset_class)

    
    for image_profile, gene_profile, positive_index, _ in loadBatchData(cur_train_data_mat, 
                                                                        cur_train_matrix_mat,
                                                                        cur_train_coors_mat, 
                                                                        params['batch_size_pair'],
                                                                        cur_pos_info):
        
        input_gene_exp = torch.tensor(np.asarray(gene_profile)).float().to(device) # torch.Size([64, 128])
        image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [1331, 256, 384] --> [1331*256, 384]
        input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
        
        ## model
        (representation_matrix, 
         reconstruction_matrix, 
         projection_matrix,
         representation_image, 
         reconstruction_iamge, 
         projection_image) = model(input_gene_exp, input_image_exp)     
        
        ## reshape
        _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
        _, projection_image_reshape = reshape_latent_image(projection_image, dataset_class)
    
        ## cross decoder
        reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
        _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)

        ## compute the loss
        loss = l(
            projection_image_reshape,
            projection_matrix,
            # representation_image_reshape,
            # representation_matrix,
            torch.tensor(positive_index).to(device),
            input_image_exp,
            reconstruction_iamge,
            reconstruction_matrix,
            reconstruction_iamge_reshapef2,
            # input_gene_exp,
            # w1=1, w2=1, w3=1, w4=1
            input_gene_exp
        )  

        # optimization
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        total_loss += loss.item() * gene_profile.shape[0]
        total_num += gene_profile.shape[0]
        pass
    
    LOSS = total_loss/total_num    # shumin end

    return LOSS



def test_model(params, model, test_loader, l, tree_type, leaf_size, dataset_class): 
    
    print("test model")
    
    # load data
    (cur_train_data_mat, 
     cur_train_matrix_mat, 
     cur_train_coors_mat, 
     cur_pos_info) = loadTrainTestData(test_loader, neighbor_k=params['k_nearest_positives'], 
                                       tree_type=tree_type, leaf_size=leaf_size, dataset_class=dataset_class)
        
    total_loss, total_num = 0.0, 0.0
    for image_profile, gene_profile, positive_index, _ in loadBatchData(cur_train_data_mat, 
                                                                        cur_train_matrix_mat,
                                                                        cur_train_coors_mat, 
                                                                        params['batch_size_pair'],
                                                                        cur_pos_info):
        
        input_gene_exp = torch.tensor(np.asarray(gene_profile)).float().to(device) # torch.Size([64, 128])
        image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [1331, 256, 384] --> [1331*256, 384]
        input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
        
        ## model
        (representation_matrix, 
         reconstruction_matrix, 
         projection_matrix,
         representation_image, 
         reconstruction_iamge, 
         projection_image) = model(input_gene_exp, input_image_exp)     
        
        ## reshape
        _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
        _, projection_image_reshape = reshape_latent_image(projection_image, dataset_class)
    
        ## cross decoder
        reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
        _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)

        ## compute the loss
        loss = l(
            projection_image_reshape,
            projection_matrix,
            # representation_image_reshape,
            # representation_matrix,
            torch.tensor(positive_index).to(device),
            input_image_exp,
            reconstruction_iamge,
            reconstruction_matrix,
            reconstruction_iamge_reshapef2,
            # input_gene_exp,
            # w1=1, w2=1, w3=1, w4=1
            input_gene_exp
        )  
    
        total_loss += loss.item() * gene_profile.shape[0]
        total_num += gene_profile.shape[0]
        pass
    
    LOSS = total_loss/total_num 
    
    # count = batch["reduced_expression"].size(0)
    # loss_meter.update(loss.item(), count)
    # tqdm_object.set_postfix(valid_loss=loss_meter.avg)

    return LOSS



##################################
# 2025.01.06 Adjust loss function
# Train and Test model adjust
##################################
# def train_model(params, model, train_loader, optimizer, cur_epoch, l, tree_type, leaf_size, dataset_class): 
    
#     print("train model")    
    
#     cur_lr = adjust_learning_rate(optimizer, cur_epoch, params['inital_learning_rate'], params['training_epoch'])
#     total_loss, total_num = 0.0, 0.0

#     ## load data
#     (cur_train_data_mat, 
#      cur_train_matrix_mat, 
#      cur_train_coors_mat, 
#      cur_pos_info) = loadTrainTestData(train_loader, neighbor_k=params['k_nearest_positives'], 
#                                        tree_type=tree_type, leaf_size=leaf_size, dataset_class=dataset_class)

    
#     for image_profile, gene_profile, positive_index, _ in loadBatchData(cur_train_data_mat, 
#                                                                         cur_train_matrix_mat,
#                                                                         cur_train_coors_mat, 
#                                                                         params['batch_size_pair'],
#                                                                         cur_pos_info):
        
#         input_gene_exp = torch.tensor(np.asarray(gene_profile)).float().to(device)   # torch.Size([64, 128])
#         image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [1331, 256, 384] --> [1331*256, 384]
#         input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
        
#         ## model
#         (representation_matrix, 
#          reconstruction_matrix, 
#          projection_matrix,
#          representation_image, 
#          reconstruction_iamge, 
#          projection_image,
#          recon_mat_mean, recon_mat_disp, recon_mat_pi) = model(input_gene_exp, input_image_exp)     
        
#         ## reshape
#         _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
#         _, projection_image_reshape = reshape_latent_image(projection_image, dataset_class)
    
#         ## cross decoder
#         reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
#         _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)

#         ## compute the loss
#         loss = l(
#             projection_image_reshape,
#             projection_matrix,
#             # representation_image_reshape,
#             # representation_matrix,
#             torch.tensor(positive_index).to(device),
#             input_image_exp,
#             reconstruction_iamge,
#             reconstruction_matrix,
#             reconstruction_iamge_reshapef2,
#             # input_gene_exp,
#             # w1=1, w2=1, w3=1, w4=1
#             input_gene_exp,
#             recon_mat_mean, recon_mat_disp, recon_mat_pi
#         )  

#         # optimization
#         optimizer.zero_grad()
#         loss.backward()
#         optimizer.step()

#         total_loss += loss.item() * gene_profile.shape[0]
#         total_num += gene_profile.shape[0]
#         pass
    
#     LOSS = total_loss/total_num    # shumin end

#     return LOSS



# def test_model(params, model, test_loader, l, tree_type, leaf_size, dataset_class): 
    
#     print("test model")
    
#     # load data
#     (cur_train_data_mat, 
#      cur_train_matrix_mat, 
#      cur_train_coors_mat, 
#      cur_pos_info) = loadTrainTestData(test_loader, neighbor_k=params['k_nearest_positives'], 
#                                        tree_type=tree_type, leaf_size=leaf_size, dataset_class=dataset_class)
        
#     total_loss, total_num = 0.0, 0.0
#     for image_profile, gene_profile, positive_index, _ in loadBatchData(cur_train_data_mat, 
#                                                                         cur_train_matrix_mat,
#                                                                         cur_train_coors_mat, 
#                                                                         params['batch_size_pair'],
#                                                                         cur_pos_info):
        
#         input_gene_exp = torch.tensor(np.asarray(gene_profile)).float().to(device)   # torch.Size([64, 128])
#         image_profile_reshape = image_profile.view(-1, image_profile.shape[2])     # [1331, 256, 384] --> [1331*256, 384]
#         input_image_exp = image_profile_reshape.clone().detach().to(device)     # SDU
        
#         ## model
#         (representation_matrix, 
#          reconstruction_matrix, 
#          projection_matrix,
#          representation_image, 
#          reconstruction_iamge, 
#          projection_image,
#          recon_mat_mean, recon_mat_disp, recon_mat_pi) = model(input_gene_exp, input_image_exp)     
        
#         ## reshape
#         _, representation_image_reshape = reshape_latent_image(representation_image, dataset_class)
#         _, projection_image_reshape = reshape_latent_image(projection_image, dataset_class)
    
#         ## cross decoder
#         reconstructed_matrix_reshaped = model.matrix_decoder(representation_image)  
#         _, reconstruction_iamge_reshapef2 = reshape_latent_image(reconstructed_matrix_reshaped, dataset_class)

#         ## compute the loss
#         loss = l(
#             projection_image_reshape,
#             projection_matrix,
#             # representation_image_reshape,
#             # representation_matrix,
#             torch.tensor(positive_index).to(device),
#             input_image_exp,
#             reconstruction_iamge,
#             reconstruction_matrix,
#             reconstruction_iamge_reshapef2,
#             # input_gene_exp,
#             # w1=1, w2=1, w3=1, w4=1
#             input_gene_exp,
#             recon_mat_mean, recon_mat_disp, recon_mat_pi
#         )  
    
#         total_loss += loss.item() * gene_profile.shape[0]
#         total_num += gene_profile.shape[0]
#         pass
    
#     LOSS = total_loss/total_num 
    
#     # count = batch["reduced_expression"].size(0)
#     # loss_meter.update(loss.item(), count)
#     # tqdm_object.set_postfix(valid_loss=loss_meter.avg)

#     return LOSS


#################################################
# 2025.01.24: add the Train model im main code
#################################################
def train_model_fst(params, model, train_loader, test_loader, 
                    optimizer, l, dir_name, logger, dataset_class='Visium64'):
    
    logger.info('Begin Training ...')
    start_train_time = time.time()

    best_loss = float('inf')
    best_epoch = 0

    ## Initialize lists to store losses
    train_losses = []
    test_losses = []

    for epoch in range(params['training_epoch']):
        logger.info('epoch [{}/{}]'.format(epoch + 1, params['training_epoch']))
        print(f"Epoch: {epoch + 1}")

        #####################
        # Train the model
        #####################
        model.train()
        start_time = time.time()
        train_loss = train_model(params, model, train_loader, optimizer, epoch, l, 
                                     tree_type='KDTree', leaf_size=2, dataset_class=dataset_class)   
        print("--- %s seconds ---" % (time.time() - start_time))
        
        ## Append training loss
        train_losses.append(train_loss)
        
        #####################
        # Evaluate the model
        #####################
        model.eval()
        with torch.no_grad():
            test_loss = test_model(params, model, test_loader, l, 
                                       tree_type='KDTree', leaf_size=2, dataset_class=dataset_class) 
            # Append test loss
            test_losses.append(test_loss)
        
        if best_loss > test_loss:
            best_loss = test_loss
            best_epoch = epoch

            save_model(model, dir_name, params, optimizer, train_loss)
            print("Saved Best epoch & Best Model! Loss: [{}: {}]".format(best_epoch, best_loss))
            logger.info("Saved Best epoch & Best Model! Loss: [{}: {}]".format(best_epoch, best_loss))

    print("Done!, final loss: {}".format(best_loss))
    print("Best epoch: {}".format(best_epoch))
    
    print("--- %s seconds ---" % (time.time() - start_train_time))
    logger.info("Done!, Best epoch & Best Model! Loss: [{}: {}]".format(best_epoch, best_loss))
    logger.info('Finished Training')

    return dir_name, train_losses, test_losses, best_epoch, best_loss


import pandas as pd
import scanpy as sc
import time
from scipy.sparse import csr_matrix
from .utils import *
from .inference import *
import pickle
import json
from pathlib import Path
import anndata
import numpy as np
from scipy.spatial import cKDTree
import torch
import pickle


###################################################
# 2025.01.16 add embeds_convert from istar
###################################################
def istar_embeds_convert(hist_emb, locs, current_shape, image_embedings='sub', k=16):
    """
    Processes the embeddings and calculates the nearest pixel locations.
    Parameters:
        hist_emb (dict): Dictionary containing the 'sub' key with the embeddings.
        locs (numpy.ndarray): The locations to be processed.
        current_shape (numpy.ndarray): The current shape.
        target_shape (numpy.ndarray): The target shape.
    Returns:
        numpy.ndarray: The ordered locations.
        numpy.ndarray: The ordered images.
    """
    def load_pickle(filename, verbose=True):
        with open(filename, 'rb') as file:
            x = pickle.load(file)
        if verbose:
            print(f'Pickle loaded from {filename}')
        return x

    ## form (W-H) to (H-W)
    locs = locs[['y','x']]    
    embs = load_pickle(str(hist_emb))

    ## Reshape the 3D array into the shape (width, height, depth)
    if image_embedings=='sub':
        embs_sub = np.array(embs['sub'])
        embs = embs_sub.transpose(1, 2, 0)
    elif image_embedings=='cls':
        embs_cls = np.array(embs['cls'])
        embs = embs_cls.transpose(1, 2, 0)
    elif image_embedings=='cls_sub':
        embs = np.concatenate([embs['cls'], embs['sub']])
        embs = embs.transpose(1, 2, 0)
    elif image_embedings=='cls_sub_rgb':
        embs = np.concatenate([embs['cls'], embs['sub'], embs['rgb']])
        embs = embs.transpose(1, 2, 0)
    
    ## Rescale locations
    target_shape = embs.shape[:2]
    rescale_factor = current_shape // target_shape
    locs = locs.astype(float)
    locs /= rescale_factor
    locs = locs.round().astype(int)

    ## convert (width, height, depth) to (width*height, depth)
    imgs_pixel = []
    locs_pixel = []
    for i in range(embs.shape[0]):        # height
        for j in range(embs.shape[1]):    # width
            embeddings = embs[i, j, :]
            if not np.isnan(embeddings).any():  # Check if any value in embeddings is NaN
                imgs_pixel.append(embeddings)
                locs_pixel.append([i+1, j+1])

    locs_pixel = np.array(locs_pixel)
    imgs_pixel = np.array(imgs_pixel)

    ## convert (width*height, depth) to (#spots, depth)
    tree = cKDTree(locs_pixel)
    _, closest_points_indices = tree.query(locs, k)

    locs_order = locs_pixel[closest_points_indices]
    imgs_order = imgs_pixel[closest_points_indices]
    imgs_order_2d = torch.from_numpy(imgs_order.reshape(imgs_order.shape[0]*imgs_order.shape[1], -1))
                                     
    return locs_order, imgs_order, imgs_order_2d


###################################################
# 2025.01.16 add patch_size calculate from adata
# Return x_dir： patch_size at x/weidth direction 
###################################################

def patch_size(adata, p=16, dir='x'):
    """
    Computes the absolute differences of the sorted spatial data in adata.
    Parameters:
        adata: anndata object which contains the spatial data
        p: int, number of rows to select after sorting (default is 16)
        dir: str, direction to sort by, either 'x' or 'y' (default is 'x')
    Returns:
        differences: pandas Series, the computed absolute differences
    """
    if dir == 'x':    # fix hight
        spatial_test = pd.DataFrame(adata.obsm['spatial']).sort_values(by=1)[:p]
        differences = spatial_test[0].diff().abs()
    elif dir == 'y':   # fix weidth
        spatial_test = pd.DataFrame(adata.obsm['spatial']).sort_values(by=0)[:p]
        differences = spatial_test[1].diff().abs()
    else:
        print("Invalid direction. Please choose either 'x' or 'y'.")

    differences = differences.dropna()
    return differences


########################################
# 2025.01.15 For using istar img feature
########################################
def position_order_adata_istar(position, obs_names, dataset_class='Visium16'):
    ## Filter rows and set new index
    position_order = position[position[position.columns[-5]] == 1]
    position_order = position_order.set_index(position_order.columns[-6])

    ## Reorder index and drop column
    position_order = position_order.reindex(obs_names)
    position_order = position_order.drop(columns=[position.columns[-5]])

    ## Rename and reorder columns
    if dataset_class == 'Visium16' or dataset_class == 'Visium64':
        position_order.columns = ['array_col', 'array_row', 'pixel_x', 'pixel_y']
    elif dataset_class == 'VisiumHD':
        position_order.columns = ['array_col', 'array_row', 'pixel_x', 'pixel_y']
    elif dataset_class == 'VisiumHD_MS64':
        position_order.columns = ['array_col', 'array_row', 'pixel_y', 'pixel_x']
    
    position_order = position_order[['pixel_y','pixel_x', 'array_row', 'array_col']]

    return position_order

########################################
# 2024.12.2 add json file
########################################
def json_load(json_path):
    path_to_visium_bundle = Path(str(json_path)).expanduser()
    with open(path_to_visium_bundle / "scalefactors_json.json") as file:
        visium_scale_factors = json.load(file)
    return visium_scale_factors

def parquet2csv(parquet_path, parquet_name='tissue_positions.parquet'):
    os.chdir(str(parquet_path))
    positions = pd.read_parquet(parquet_name)    
    positions.set_index('barcode', inplace=True)
    ## inverse pxl_row_in_fullres and pxl_col_in_fullres
    positions.columns = ['in_tissue', 'array_row', 'array_col', 'pxl_col_in_fullres', 'pxl_row_in_fullres']
    position_tissue = positions[positions['in_tissue'] == 1]
    return position_tissue


######################################
# 2024.11.13 add for Visium position
######################################
def filter_pos_list(filename):
    """
    Reads CSV file, renames the columns, and filters the rows where 'in_tissue' is 1.
        filename: str, The name of the CSV file to read.
        Returns a DataFrame.
    """
    position = pd.read_csv(filename, header=None)
    position.columns = ['barcode', 'in_tissue', 'array_row', 'array_col', 
                        'pxl_row_in_fullres', 'pxl_col_in_fullres']
    position_tissue = position[position['in_tissue'] == 1]
    return position_tissue


######################################
# 2024.11.13 add for Interpolate spot
# faster version
######################################
def inter_spot(position, direction):
    """
    Returns a DataFrame of midpoints between adjacent points in the specified direction.
        position : DataFrame
        direction: str, either 'x' or 'y'
    """
    ## Order position according to array_row and array_col 
    position_ordered = position.sort_values(['array_col', 'array_row'], 
                                            ascending=[True, True]).reset_index(drop=True)

    mid_points_list = []

    ## Iterate over position_ordered
    for _, row in position_ordered.iterrows():
        if direction == 'x':
            # Find the next point in the 'x' direction
            next_row = position_ordered[
                (position_ordered['array_row'] == row['array_row'] + 2) & 
                (position_ordered['array_col'] == row['array_col'])
            ]

            if not next_row.empty:
                next_row = next_row.iloc[0]
                mid_pxl_row = (row['pxl_row_in_fullres'] + next_row['pxl_row_in_fullres']) / 2
                mid_pxl_col = (row['pxl_col_in_fullres'] + next_row['pxl_col_in_fullres']) / 2

                mid_points_list.append({
                    'array_row': row['array_row'] + 1,
                    'array_col': row['array_col'],
                    'pxl_row_in_fullres': mid_pxl_row,
                    'pxl_col_in_fullres': mid_pxl_col
                })

        elif direction == 'y' and row['array_col'] < 127:
            ## Find the nearest points in the horizontal direction
            nearest_rows = position_ordered[
                position_ordered['array_col'] == row['array_col'] + 1
            ].copy()
            nearest_rows['distance'] = np.abs(nearest_rows['array_row'] - row['array_row'])
            nearest_rows = nearest_rows.nsmallest(2, 'distance')

            ## Compute the midpoints
            for _, nearest_row in nearest_rows.iterrows():
                mid_points_list.append({
                    'array_row': (row['array_row'] + nearest_row['array_row']) / 2,
                    'array_col': (row['array_col'] + nearest_row['array_col']) / 2,
                    'pxl_row_in_fullres': (row['pxl_row_in_fullres'] + nearest_row['pxl_row_in_fullres']) / 2,
                    'pxl_col_in_fullres': (row['pxl_col_in_fullres'] + nearest_row['pxl_col_in_fullres']) / 2
                })

    ## Create DataFrame from list of midpoints
    position_add = pd.DataFrame(mid_points_list)

    if direction == 'y':
        position_add = position_add.drop_duplicates()    # Remove duplicates for 'y' direction

    return position_add


def final_pos_list(position_x, position_y, position=None):
    """
    Returns DataFrame: the concatenation of position, 
    position_x, and position_y, sorted by 'array_col' and 'array_row'.
        position_x, position_y : DataFrame
    """
    ## Concatenate position (if provided), position_x, and position_y
    if position is not None:
        position_final = pd.concat([
            position[['array_row', 'array_col', 
                      'pxl_row_in_fullres', 'pxl_col_in_fullres']], 
            position_x, 
            position_y
        ], ignore_index=True)
    else:
        position_final = pd.concat([position_x, position_y], ignore_index=True)

    # #Sort position_final by 'array_col' and 'array_row'
    position_final = position_final.sort_values(['array_col', 'array_row'], 
                                                ascending=[True, True]).reset_index(drop=True)

    return position_final


######################################
# 2024.11.12 add for pathway analysis
######################################
def clean_save_adata(adata, filename):
    adata_save = adata.copy()

    ## List of keys to remove
    keys_to_remove = ['single_cell', 'mean', 'num_pairs',
                      # 'ligand', 'receptor',  'geneInter'                      
                      'global_I', 'global_stat', 'global_res', 'local_z'
                      # 'local_stat', 'local_z_p', 
                      # 'selected_spots',
                      ]

    for key in keys_to_remove:
        if key in adata_save.uns:
            del adata_save.uns[key]

    ## Update problematic elements in adata_save.uns and save them as pickle files
    for key, value in adata_save.uns.items():
        ## Save the original value as a pickle file
        with open(f"{key}.pkl", "wb") as f:
            pickle.dump(value, f)
        ## Update the value in adata with the filename of the pickle file
        adata_save.uns[key] = f"{key}.pkl"

    adata_save.write_h5ad(filename)
    return adata_save


def Load_clean_save_adata(adata):
    keys = ["local_z_p", "local_stat", "geneInter", "ligand", "receptor", "selected_spots",
            'histology_results_binary', 'histology_results_continu']
    for key in keys:
        with open(adata.uns[key], "rb") as f:
            adata.uns[key] = pickle.load(f)
    return adata


######################################
# 2024.11.11 add for all spot: Visium
######################################
def get_allspot_coors(input_coord_all):
    
    tensor_1 = input_coord_all[0][0]
    tensor_2 = input_coord_all[0][1]

    input_coord_all_concat = torch.stack((tensor_1, tensor_2))
    spatial_loc = input_coord_all_concat.T.numpy()

    ## Find unique rows and their counts
    unique_rows, counts = np.unique(spatial_loc, axis=0, return_counts=True)
    ## Check if there are any duplicate rows
    duplicate_rows = (counts > 1).any()
    print("Are there any duplicate rows? :", duplicate_rows)
    return spatial_loc


def adata_LR(adata, file_path):
    LRgene = pd.read_csv(file_path)
    adata.var_names_make_unique()
    if isinstance(adata.X, np.ndarray):
        adata_matrix = pd.DataFrame(adata.X, index=adata.obs_names, columns=adata.var_names)
    else:
        adata_matrix = pd.DataFrame(adata.X.A, index=adata.obs_names, columns=adata.var_names)
    available_genes = [gene for gene in LRgene['LR gene'].tolist() if gene in adata_matrix.columns]
    adataLR_matrix = adata_matrix[available_genes]
    adata._n_vars = adataLR_matrix.shape[1]
    adata.X = adataLR_matrix.values
    adata.var = adata.var.loc[available_genes]
    adata.var_names = adataLR_matrix.columns
    return adata


#####################################
# 2024.12.23: Add 
#####################################
def adata_preprocess(adata, keep_raw=False, normalize=True, 
                     min_cells=10, target_sum=None, n_top_genes=None, species='human'):
    """
    Preprocesses AnnData object for single-cell RNA-seq data.
    Parameters:
        adata (anndata.AnnData): The annotated data matrix of shape n_obs x n_vars. 
        keep_raw (bool, optional): If True, a copy of the original data is saved. Default is False.
        min_cells (int, optional): Minimum number of cells expressed. Default is 10.
        target_sum (float, optional): If not None, normalize total counts per cell with this value. 
                                    If None, after normalization, each cell has a total count 
                                    equal to the median of the counts_per_cell before normalization. 
                                    Default is None.
        n_top_genes (int, optional): Number of highly-variable genes to keep. 
                                    If n_top_genes is not None, this number is kept as 
                                    highly-variable genes. Default is None.
        species (str, optional): The species of the dataset. If not 'human', certain steps are skipped.
    Returns:
        adata (anndata.AnnData): The processed annotated data matrix.
    """

    ## Set mitochondrial gene prefix based on species
    if species == 'human':
        adata.var["mt"] = adata.var_names.str.startswith("MT-")
    elif species == 'mouse':
        adata.var["mt"] = adata.var_names.str.startswith("mt-")
    else:
        raise ValueError("Unsupported species. Please specify 'human' or 'mouse'.")

    ## Calculate QC metrics if there are mitochondrial genes
    if adata.var["mt"].any():
        sc.pp.calculate_qc_metrics(adata, qc_vars=["mt"], inplace=True)
        
    sc.pp.filter_genes(adata, min_cells=min_cells)

    if keep_raw:
        adata = adata.copy()     # del adata.raw   

    if normalize:
        if target_sum is not None:
            sc.pp.normalize_total(adata, target_sum=target_sum)
        else:
            sc.pp.normalize_total(adata)

        sc.pp.log1p(adata)

    if n_top_genes is not None:
        sc.pp.highly_variable_genes(adata, flavor="seurat", n_top_genes=n_top_genes)

    return adata


def adata2matrix(adata, gene_hv):
    ## Access the matrix and convert it to a dense matrix
    if isinstance(adata.X, np.ndarray):
        matrix = pd.DataFrame(adata.X)
    else:
        matrix = pd.DataFrame(adata.X.todense())
    matrix.columns = gene_hv
    spotID = np.array(pd.DataFrame(adata.obs['in_tissue']).index)
    matrix.insert(0, '', spotID)   
    matrix = matrix.set_index(matrix.columns[0])
    print(matrix.shape)
    return matrix

#
###############################################
# 2024.11.02 update 
# 2024.12.18 add VisiumSC
###############################################
def get_image_coord(file_paths, dataset_class):
    data = []
    file_paths.sort() 
    for file_path in file_paths:
        parts = file_path.split('_')
        if dataset_class == 'Visium' or dataset_class == 'VisiumSC':
            part_3 = int(parts[-2])
            part_4 = int(parts[-1].split('.')[0])
        elif dataset_class == 'VisiumHD':
            part_3 = parts[-2]
            part_4 = parts[-1].split('.pth')[0]
        else:
            print("Invalid dataset_class. Please use 'Visium', 'VisiumSC' or 'VisiumHD'")
            return
        data.append([part_3, part_4])
    df = pd.DataFrame(data, columns=['pixel_y', 'pixel_x'])
    return df[['pixel_x', 'pixel_y']]
    

def get_image_coord_all(file_paths):
    file_paths.sort()
    data = []
    for file_path in file_paths:
        parts = file_path.split('_')
        data.append([parts[-2], parts[-1].split('.pth')[0]])
        # if dataset_class == 'Visium' or dataset_class == 'VisiumSC':
        #     data.append([parts[-2], parts[-1].split('.pth')[0]])
    return data


def image_coord_merge(df, position, dataset_class):
    def merge_dfs(df, position):
        merged_df = pd.merge(df, position, on=['pixel_x', 'pixel_y'], how='left')
        cols = merged_df.columns.tolist()
        cols.remove('pixel_x')
        cols.remove('pixel_y')
        merged_df = merged_df[cols + ['pixel_x', 'pixel_y']]
        col_x = merged_df.columns[-4]
        col_y = merged_df.columns[-3]
        return merged_df.rename(columns={col_x: 'x', col_y: 'y'})

    #######################################################
    # 2024.12.04 postion doesn't match image
    #######################################################
    def merge_dfs_HD(df, position):
        position['pxl_col_in_fullres'] = pd.to_numeric(position['pxl_col_in_fullres'], errors='coerce').round(6)
        position['pxl_row_in_fullres'] = pd.to_numeric(position['pxl_row_in_fullres'], errors='coerce').round(6)
        position = position.rename(columns={'pxl_col_in_fullres': 'pixel_x', 'pxl_row_in_fullres': 'pixel_y'})

        df['pixel_x'] = df['pixel_x'].astype('float64').round(6)
        df['pixel_y'] = df['pixel_y'].astype('float64').round(6)

        merged_df = pd.merge(df, position, on=['pixel_x', 'pixel_y'], how='left')
        cols = merged_df.columns.tolist()
        cols.remove('pixel_x')
        cols.remove('pixel_y')
        merged_df = merged_df[cols + ['pixel_x', 'pixel_y']]
        col_x = merged_df.columns[-4]
        col_y = merged_df.columns[-3]
        return merged_df.rename(columns={col_x: 'x', col_y: 'y'})

    #     in_df = position['pixel_x'].isin(df['pixel_x']) & position['pixel_y'].isin(df['pixel_y'])
    #     merged_df = position[in_df].reset_index(drop=True)
    #     merged_df = merged_df.rename(columns={'array_row': 'x', 'array_col': 'y'})
    #     return merged_df

    ## Use dataset_class to decide which function to call
    if dataset_class == 'Visium' or dataset_class=='VisiumSC':
        result = merge_dfs(df, position)
    elif dataset_class == 'VisiumHD':
        result = merge_dfs_HD(df, position)
    else:
        raise ValueError(f"Unknown dataset_class: {dataset_class}")

    ## Check if the merge was successful
    if result.empty:
        raise ValueError("The merging resulted in an empty DataFrame. Please check your input data.")

    # ## For Visium
    # result = result.drop_duplicates(subset=result.columns[0], keep='first')

    return result


def sort_matrix(matrix, position_image, spotID_order, gene_hv):
    ## Reset the index of the matrix and rename the first column
    position_image_first_col = position_image.columns[0]
    matrix = matrix.reset_index().rename(columns={matrix.index.name: position_image_first_col})
    
    ## Merge position_image and matrix based on the first column
    sorted_matrix = pd.merge(position_image[[position_image_first_col]], matrix, 
                             on=position_image_first_col, how="left")
    
    # ################################################
    # # different: delete the same row For VisiumHD
    # ################################################
    # sorted_matrix = sorted_matrix.drop_duplicates(subset=position_image_first_col, keep='first')
    
    matrix_order = np.array(sorted_matrix.set_index(position_image_first_col))
    
    ## Convert matrix_order to DataFrame and set the index and column names
    matrix_order_df = pd.DataFrame(matrix_order)
    matrix_order_df.index = spotID_order
    matrix_order_df.columns = gene_hv
    
    return matrix_order, matrix_order_df


def update_adata_coord(adata, matrix_order, position_image, 
                       spotID_order=None, gene_hv=None, dataset_class='Visium16'):
    if dataset_class in ['Visium16', 'Visium64']:
        adata.X = csr_matrix(matrix_order, dtype=np.float32)
        adata.obs_names = matrix_order.index    # order by image feature name 
        adata.obsm['spatial'] = np.array(position_image.loc[:, ['pixel_y', 'pixel_x']])
        adata.obs['array_row'] = np.array(position_image.loc[:, 'y'])
        adata.obs['array_col'] = np.array(position_image.loc[:, 'x'])

    elif dataset_class == 'VisiumHD':
        sparse_matrix = csr_matrix(matrix_order, dtype=np.float32)
        ## construct new adata (reduce 97 coords)
        adata_redu = sc.AnnData(X=sparse_matrix, 
                                obs=pd.DataFrame(index=spotID_order), 
                                var=pd.DataFrame(index=gene_hv))
        adata_redu.X = csr_matrix(matrix_order, dtype=np.float32)
        adata_redu.obsm['spatial'] = np.array(position_image.loc[:, ['pixel_y', 'pixel_x']])
        adata_redu.obs['array_row'] = np.array(position_image.loc[:, 'y'])
        adata_redu.obs['array_col'] = np.array(position_image.loc[:, 'x'])
        ## add another objects
        adata_redu.var = adata.var
        adata_redu.uns = adata.uns
        adata = adata_redu.copy()
    else:
        raise ValueError("Invalid dataset_class. Expected 'Visium16', 'Visium64' or 'VisiumHD'.")
    
    return adata


def update_st_coord(position_image):
    position_order = pd.DataFrame({
        "pixel_y": position_image.loc[:, 'pixel_y'],
        "pixel_x": position_image.loc[:, 'pixel_x'],
        "array_row": position_image.loc[:, 'y'],
        "array_col": position_image.loc[:, 'x']
    })
    return position_order


###########################################
# 2025.01.06 sdjust weight and optimal nbs
###########################################
def impute_adata(adata, adata_spot, C2, gene_hv, dataset_class, weight_exponent=1):
    '''
    Prepare impute_adata: Fill gene expression using nbs
        adata_know: adata (original) 1331 × 596
        adata_spot: all subspot 21296 × 596
    '''
    adata_know = adata.copy()
    adata_know.obs[["x", "y"]] = adata.obsm['spatial']
    # adata_spot.obsm['spatial'] = adata_spot.obs[["x", "y"]].values
    sudo = pd.DataFrame(C2, columns=["x", "y"])
    sudo_adata = sc.AnnData(np.zeros((sudo.shape[0], len(gene_hv))), obs=sudo, var=adata.var)

    ## set ‘split_num’, according 'dataset_class'
    if dataset_class == 'Visium16':
        k_nbs = 6
        split_num = 16
    elif dataset_class == 'Visium64':
        k_nbs = 6
        split_num = 64
    elif dataset_class == 'VisiumSC':
        k_nbs = 6
        split_num = 1
    elif dataset_class == 'VisiumHD':
        k_nbs = 4
        split_num = 4
    else:
        raise ValueError('Invalid dataset_class. Only "Visium16", "Visium64", "VisiumSC" and "VisiumHD" are supported.')

    ## Impute_adata
    start_time = time.time()

    nearest_points = find_nearest_point(adata_spot.obsm['spatial'], adata_know.obsm['spatial'])
    nbs, nbs_indices = find_nearest_neighbors(nearest_points, adata_know.obsm['spatial'], k=k_nbs)
    distances = calculate_euclidean_distances(adata_spot.obsm['spatial'], nbs)

    ## Iterate over each point in sudo_adata
    for i in range(sudo_adata.shape[0]):
        dis_tmp = (distances[i] + 0.1) / np.min(distances[i] + 0.1)
        weights = ((1 / (dis_tmp ** weight_exponent)) / ((1 / (dis_tmp ** weight_exponent)).sum()))
        
        if isinstance(adata_know.X, np.ndarray):
            sudo_adata.X[i, :] = np.dot(weights, adata_know.X[nbs_indices[i]])
        else:
            sudo_adata.X[i, :] = np.dot(weights, adata_know.X[nbs_indices[i]].todense())

        # sudo_adata.X[i, :] = np.dot(weights, adata_know.X[nbs_indices[i]].todense())
        # sudo_adata.X[i, :] = np.dot(weights, adata_know.X[nbs_indices[i]].todense() / split_num)

    print("--- %s seconds ---" % (time.time() - start_time))

    ## add other objects to adata
    sudo_adata.obsm['spatial'] = adata_spot.obsm['spatial']
    sudo_adata.uns['spatial'] = adata.uns['spatial']

    return sudo_adata

######################################
# 2025.02.08: add scale 
######################################
def weight_adata(adata_spot, sudo_adata, gene_hv, w=0.5, do_scale=False):
    """
    Combine inferred super-resolved gene expression data with imputed data, and optionally scale the result.
    Parameters:
        adata_spot (sc.AnnData): Inferred super-resolved gene expression data with high resolution.
        sudo_adata (sc.AnnData): Imputed data using k-nearest neighbors within spots.
        gene_hv (list): List of highly variable genes.
        w (float, optional): Weight for combining the two datasets. Defaults to 0.5.
        do_scale (bool, optional): Whether to scale the combined data. Defaults to False.
    Returns:
        sc.AnnData: Combined and optionally scaled AnnData object.
        torch.Tensor: The combined data as a PyTorch tensor.
    """

    ## Optionally scale the combined data
    if do_scale:
        weight_impt_data = w * scale(adata_spot.X) + (1 - w) * scale(sudo_adata.X)
    else:
        weight_impt_data = w * adata_spot.X + (1 - w) * sudo_adata.X

    ## Convert the combined data to a PyTorch tensor
    data_impt = torch.tensor(weight_impt_data)

    ## Create a new AnnData object with the combined data
    adata_impt = sc.AnnData(X=pd.DataFrame(weight_impt_data))
    adata_impt.var_names = gene_hv
    adata_impt.obs = adata_spot.obs

    ## Add other necessary objects to the new AnnData object
    adata_impt.obsm['spatial'] = sudo_adata.obsm['spatial']
    adata_impt.uns['spatial'] = sudo_adata.uns['spatial']

    return adata_impt, data_impt


def reshape2adata(adata, adata_impt_all_reshape, gene_hv, spatial_loc_all=None):

    if isinstance(adata_impt_all_reshape, torch.Tensor):
        adata_impt_spot = sc.AnnData(X = adata_impt_all_reshape.numpy())
    elif isinstance(adata_impt_all_reshape, anndata.AnnData):
        adata_impt_spot = sc.AnnData(X = adata_impt_all_reshape.to_df())

    adata_impt_spot.var_names = gene_hv

    if adata_impt_all_reshape.shape[0] == adata.shape[0]:
        adata_impt_spot.obs_names = adata.obs_names
        adata_impt_spot.obsm['spatial'] = adata.obsm['spatial'] 
        adata_impt_spot.obs = adata.obs
        adata_impt_spot.var = adata.var
    else: 
        adata_impt_spot.obs['x'] = spatial_loc_all[:,0]
        adata_impt_spot.obs['y'] = spatial_loc_all[:,1]
        adata_impt_spot.obsm['spatial'] = adata_impt_spot.obs[["x", "y"]].values
    
    ## Check if adata has 'uns' and 'spatial' key
    if hasattr(adata, 'uns') and 'spatial' in adata.uns:
        adata_impt_spot.uns['spatial'] = adata.uns['spatial']

    return adata_impt_spot